@@ -34,9 +34,6 @@
 #include "optscript.h"
 #include "script_p.h"
 
-#define FIELD_NULL_LETTER_CHAR '-'
-#define FIELD_NULL_LETTER_STRING "-"
-
 typedef struct sFieldObject {
 	fieldDefinition *def;
 	vString     *buffer;
@@ -78,22 +75,23 @@ static bool doesContainAnyCharInInput (const tagEntryInfo *const tag, const char
 static bool doesContainAnyCharInFieldScope (const tagEntryInfo *const tag, const char *value, const char *chars);
 static bool doesContainAnyCharInSignature (const tagEntryInfo *const tag, const char *value, const char *chars);
 
-static bool     isTyperefFieldAvailable   (const tagEntryInfo *const tag);
-static bool     isFileFieldAvailable      (const tagEntryInfo *const tag);
-static bool     isInheritsFieldAvailable  (const tagEntryInfo *const tag);
-static bool     isAccessFieldAvailable    (const tagEntryInfo *const tag);
-static bool     isImplementationFieldAvailable (const tagEntryInfo *const tag);
-static bool     isSignatureFieldAvailable (const tagEntryInfo *const tag);
-static bool     isExtrasFieldAvailable    (const tagEntryInfo *const tag);
-static bool     isXpathFieldAvailable     (const tagEntryInfo *const tag);
-static bool     isEndFieldAvailable       (const tagEntryInfo *const tag);
-static bool     isEpochAvailable          (const tagEntryInfo *const tag);
-static bool     isNthAvailable            (const tagEntryInfo *const tag);
+static bool     isTyperefFieldAvailable   (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isFileFieldAvailable      (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isInheritsFieldAvailable  (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isAccessFieldAvailable    (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isImplementationFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isSignatureFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isExtrasFieldAvailable    (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isXpathFieldAvailable     (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isEndFieldAvailable       (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isEpochAvailable          (const tagEntryInfo *const tag, const fieldDefinition *fdef);
+static bool     isNthAvailable            (const tagEntryInfo *const tag, const fieldDefinition *fdef);
 
 static EsObject* getFieldValueForName (const tagEntryInfo *, const fieldDefinition *);
 static EsObject* setFieldValueForName (tagEntryInfo *, const fieldDefinition *, const EsObject *);
 static EsObject* getFieldValueForInput (const tagEntryInfo *, const fieldDefinition *);
 static EsObject* getFieldValueForKind (const tagEntryInfo *, const fieldDefinition *);
+static EsObject* getFieldValueForLanguage (const tagEntryInfo *, const fieldDefinition *);
 static EsObject* getFieldValueForTyperef (const tagEntryInfo *, const fieldDefinition *);
 static EsObject* setFieldValueForTyperef (tagEntryInfo *, const fieldDefinition *, const EsObject *);
 static EsObject* checkFieldValueForTyperef (const fieldDefinition *, const EsObject *);
@@ -213,6 +211,11 @@ static fieldDefinition fieldDefinitionsExuberant [] = {
 		.renderNoEscaping   = NULL,
 		.doesContainAnyChar = NULL,
 		.dataType           = FIELDTYPE_STRING,
+		.getterValueType    = "language",
+		.getValueObject     = getFieldValueForLanguage,
+		.setterValueType    = NULL,
+		.checkValueForSetter= NULL,
+		.setValueObject     = NULL,
 	},
 	[FIELD_LINE_NUMBER - FIELD_ECTAGS_START] = {
 		.letter             = 'n',
@@ -226,7 +229,7 @@ static fieldDefinition fieldDefinitionsExuberant [] = {
 		.dataType           = FIELDTYPE_INTEGER,
 		.getterValueType    = "int",
 		.getValueObject     = getFieldValueForLineCommon,
-		.setterValueType    = "matchloc|int",
+		.setterValueType    = "matchloc|line:int", /* line <= getInputLineNumber(); */
 		.checkValueForSetter= checkFieldValueForLineCommon,
 		.setValueObject     = setFieldValueForLineCommon,
 	},
@@ -541,7 +544,7 @@ static fieldObject* getFieldObject(fieldType type)
 	return fieldObjects + type;
 }
 
-extern fieldType getFieldTypeForOption (char letter)
+extern fieldType getFieldTypeForLetter (char letter)
 {
 	unsigned int i;
 
@@ -617,7 +620,8 @@ extern unsigned char getFieldLetter (fieldType type)
 extern bool doesFieldHaveValue (fieldType type, const tagEntryInfo *tag)
 {
 	if (getFieldObject(type)->def->isValueAvailable)
-		return getFieldObject(type)->def->isValueAvailable(tag);
+		return getFieldObject(type)->def->isValueAvailable(tag,
+														   getFieldObject(type)->def);
 	else
 		return true;
 }
@@ -863,13 +867,12 @@ extern bool  doesFieldHaveTabOrNewlineChar (fieldType type, const tagEntryInfo *
 static const char* renderCompactInputLine (vString *b,  const char *const line)
 {
 	bool lineStarted = false;
-	const char *p;
-	int c;
 
 	/*  Write everything up to, but not including, the newline.
 	 */
-	for (p = line, c = *p  ;  c != NEWLINE  &&  c != '\0'  ;  c = *++p)
+	for (const char *p = line; *p != '\n' && *p != '\0'; ++p)
 	{
+		int c = (unsigned char) *p;
 		if (lineStarted  || ! isspace (c))  /* ignore leading spaces */
 		{
 			lineStarted = true;
@@ -879,11 +882,11 @@ static const char* renderCompactInputLine (vString *b,  const char *const line)
 
 				/*  Consume repeating white space.
 				 */
-				while (next = *(p+1) , isspace (next)  &&  next != NEWLINE)
+				while (next = (unsigned char) *(p+1), isspace (next) && next != '\n')
 					++p;
 				c = ' ';  /* force space character for any white space */
 			}
-			if (c != CRETURN  ||  *(p + 1) != NEWLINE)
+			if (c != '\r'  ||  *(p + 1) != '\n')
 				vStringPut (b, c);
 		}
 	}
@@ -1072,9 +1075,9 @@ static const char *renderFieldExtras (const tagEntryInfo *const tag,
 				     const char *value CTAGS_ATTR_UNUSED,
 				     vString* b)
 {
-	int i;
+	unsigned int i;
 	bool hasExtra = false;
-	int c = countXtags();
+	unsigned int c = countXtags();
 
 	for (i = 0; i < c; i++)
 	{
@@ -1127,9 +1130,9 @@ static const char *renderFieldEnd (const tagEntryInfo *const tag,
 {
 	static char buf[21];
 
-	if (tag->extensionFields.endLine != 0)
+	if (getTagEndLine (tag) != 0)
 	{
-		sprintf (buf, "%lu", tag->extensionFields.endLine);
+		sprintf (buf, "%lu", getTagEndLine (tag));
 		return renderAsIs (b, buf);
 	}
 	else
@@ -1141,11 +1144,16 @@ static const char *renderFieldEpoch (const tagEntryInfo *const tag,
 {
 #define buf_len 21
 	static char buf[buf_len];
-
-	if (snprintf (buf, buf_len, "%lld", (long long)tag->extensionFields.epoch) > 0)
+#ifdef _MSC_VER
+#define FMTLL "%I64d"
+#else
+#define FMTLL "%lld"
+#endif
+	if (snprintf (buf, buf_len, FMTLL, (long long)tag->extensionFields.epoch) > 0)
 		return renderAsIs (b, buf);
 	else
 		return NULL;
+#undef FMTLL
 #undef buf_len
 }
 
@@ -1163,51 +1171,43 @@ static const char *renderFieldNth (const tagEntryInfo *const tag,
 #undef buf_len
 }
 
-static bool     isTyperefFieldAvailable  (const tagEntryInfo *const tag)
+static bool isTyperefFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return (tag->extensionFields.typeRef [0] != NULL
 		&& tag->extensionFields.typeRef [1] != NULL)? true: false;
 }
 
-static bool     isFileFieldAvailable  (const tagEntryInfo *const tag)
+static bool isFileFieldAvailable  (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return tag->isFileScope? true: false;
 }
 
-static bool     isInheritsFieldAvailable (const tagEntryInfo *const tag)
+static bool isInheritsFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return (tag->extensionFields.inheritance != NULL)? true: false;
 }
 
-static bool     isAccessFieldAvailable   (const tagEntryInfo *const tag)
+static bool isAccessFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return (tag->extensionFields.access != NULL)? true: false;
 }
 
-static bool     isImplementationFieldAvailable (const tagEntryInfo *const tag)
+static bool isImplementationFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return (tag->extensionFields.implementation != NULL)? true: false;
 }
 
-static bool     isSignatureFieldAvailable (const tagEntryInfo *const tag)
+static bool isSignatureFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return (tag->extensionFields.signature != NULL)? true: false;
 }
 
-static bool     isExtrasFieldAvailable     (const tagEntryInfo *const tag)
+static bool isExtrasFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
-	unsigned int i;
-
-	if (tag->extraDynamic)
-		return true;
-	for (i = 0; i < sizeof (tag->extra); i++)
-		if (tag->extra [i])
-			return true;
-
-	return false;
+	return isTagExtra (tag);
 }
 
-static bool     isXpathFieldAvailable      (const tagEntryInfo *const tag)
+static bool isXpathFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 #ifdef HAVE_LIBXML
 	return (tag->extensionFields.xpath != NULL)? true: false;
@@ -1216,19 +1216,19 @@ static bool     isXpathFieldAvailable      (const tagEntryInfo *const tag)
 #endif
 }
 
-static bool     isEndFieldAvailable       (const tagEntryInfo *const tag)
+static bool isEndFieldAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
-	return (tag->extensionFields.endLine != 0)? true: false;
+	return (getTagEndLine(tag) != 0)? true: false;
 }
 
-static bool isEpochAvailable (const tagEntryInfo *const tag)
+static bool isEpochAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	return (tag->kindIndex == KIND_FILE_INDEX)
 		? true
 		: false;
 }
 
-static bool isNthAvailable (const tagEntryInfo *const tag)
+static bool isNthAvailable (const tagEntryInfo *const tag, const fieldDefinition *fdef CTAGS_ATTR_UNUSED)
 {
 	Assert (tag->langType >= NO_NTH_FIELD);
 	return (tag->extensionFields.nth != NO_NTH_FIELD)? true: false;
@@ -1252,7 +1252,7 @@ extern bool enableField (fieldType type, bool state)
 	else
 		verbose ("enable field \"%s\"<%s>: %s\n",
 				 getFieldObject(type)->def->name,
-				 getLanguageName (getFieldOwner(type)),
+				 getLanguageName (getFieldLanguage(type)),
 				 (state? "yes": "no"));
 	return old;
 }
@@ -1262,7 +1262,7 @@ extern bool isCommonField (fieldType type)
 	return (FIELD_BUILTIN_LAST < type)? false: true;
 }
 
-extern int     getFieldOwner (fieldType type)
+extern langType getFieldLanguage (fieldType type)
 {
 	return getFieldObject(type)->language;
 }
@@ -1285,7 +1285,7 @@ extern bool doesFieldHaveRenderer (fieldType type, bool noEscaping)
 		return getFieldObject(type)->def->render? true: false;
 }
 
-extern int countFields (void)
+extern unsigned int countFields (void)
 {
 	return fieldObjectUsed;
 }
@@ -1322,6 +1322,11 @@ static const char* defaultRenderer (const tagEntryInfo *const tag CTAGS_ATTR_UNU
 	return renderEscapedString (value, tag, buffer);
 }
 
+extern bool isValueAvailableGeneric (const tagEntryInfo *const e, const fieldDefinition *fdef)
+{
+	return getParserFieldValueForType(e, fdef->ftype)? true: false;
+}
+
 extern int defineField (fieldDefinition *def, langType language)
 {
 	fieldObject *fobj;
@@ -1332,7 +1337,7 @@ extern int defineField (fieldDefinition *def, langType language)
 	Assert (def->name);
 	for (i = 0; i < strlen (def->name); i++)
 	{
-		Assert ( isalpha (def->name [i]) );
+		Assert ( isalpha ((unsigned char) def->name [i]) );
 	}
 	def->letter = NUL_FIELD_LETTER;
 
@@ -1354,6 +1359,9 @@ extern int defineField (fieldDefinition *def, langType language)
 	if (! def->dataType)
 		def->dataType = FIELDTYPE_STRING;
 
+	if (def->isValueAvailable == NULL)
+		def->isValueAvailable = isValueAvailableGeneric;
+
 	fobj->def = def;
 
 	fobj->buffer = NULL;
@@ -1369,6 +1377,14 @@ extern int defineField (fieldDefinition *def, langType language)
 	fobj->sibling  = FIELD_UNKNOWN;
 
 	updateSiblingField (def->ftype, def->name);
+	installOptscriptFieldAccessor (def->ftype);
+
+	if (def->version > getLanguageVersionCurrent (language))
+		error (WARNING, "the version number (%u) of field \"%s\" of language \"%s\" "
+			   "should be less than or equal to the current number (%u) of the language",
+			   def->version, def->name, getLanguageName (language),
+			   getLanguageVersionCurrent (language));
+
 	return def->ftype;
 }
 
@@ -1385,7 +1401,7 @@ extern struct colprintTable * fieldColprintTableNew (void)
 {
 	return colprintTableNew ("L:LETTER", "L:NAME", "L:ENABLED",
 							 "L:LANGUAGE", "L:JSTYPE", "L:FIXED",
-							 "L:OP", "L:DESCRIPTION", NULL);
+							 "L:OP", "R:VER", "L:DESCRIPTION", NULL);
 }
 
 static void  fieldColprintAddLine (struct colprintTable *table, int i)
@@ -1416,7 +1432,7 @@ static void  fieldColprintAddLine (struct colprintTable *table, int i)
 			 bmask < FIELDTYPE_END_MARKER;
 			 bmask <<= 1, offset++)
 			if (type & bmask)
-				typefields[offset] = fieldDataTypeFalgs[offset];
+				typefields[offset] = fieldDataTypeFlags[offset];
 	}
 	colprintLineAppendColumnCString (line, typefields);
 	colprintLineAppendColumnBool (line, writerDoesTreatFieldAsFixed (i));
@@ -1427,6 +1443,7 @@ static void  fieldColprintAddLine (struct colprintTable *table, int i)
 	if (fdef->setValueObject)
 		operator[1] = 'w';
 	colprintLineAppendColumnCString (line, operator);
+	colprintLineAppendColumnVersion (line, fdef->version);
 	colprintLineAppendColumnCString (line, fdef->description);
 }
 
@@ -1604,6 +1621,12 @@ static EsObject* getFieldValueForKind (const tagEntryInfo *tag, const fieldDefin
 	return opt_name_new_from_cstr (kind_name);
 }
 
+static EsObject* getFieldValueForLanguage (const tagEntryInfo *tag, const fieldDefinition *fdef)
+{
+	const char * lang_name = getLanguageName (tag->langType);
+	return opt_name_new_from_cstr (lang_name);
+}
+
 static EsObject* getFieldValueForTyperef (const tagEntryInfo *tag, const fieldDefinition *fdef)
 {
 	if (tag->extensionFields.typeRef [0] == NULL
@@ -1713,8 +1736,12 @@ static EsObject* checkFieldValueForTyperef (const fieldDefinition *fdef, const E
 		;
 	else if (es_integer_p (obj))
 	{
-		int index = es_integer_get (obj);
-		if (index >= countEntryInCorkQueue ())
+		int index0 = es_integer_get (obj);
+		if (index0 < 0)
+			return OPT_ERR_RANGECHECK;
+
+		unsigned int index = index0;
+		if (index == 0 || index >= countEntryInCorkQueue ())
 			return OPTSCRIPT_ERR_NOTAGENTRY;
 	}
 	else
@@ -1729,10 +1756,10 @@ static EsObject* getFieldValueForScope (const tagEntryInfo *tag, const fieldDefi
 
 static EsObject* setFieldValueForScope (tagEntryInfo *tag, const fieldDefinition *fdef, const EsObject *obj)
 {
-	int index = es_integer_get (obj);
+	unsigned int index = es_integer_get (obj);
 	if (index < countEntryInCorkQueue ())
 	{
-		tag->extensionFields.scopeIndex = index;
+		tag->extensionFields.scopeIndex = (int)index;
 		return es_false;
 	}
 
@@ -1757,12 +1784,12 @@ static EsObject* getFieldValueForExtras (const tagEntryInfo *tag, const fieldDef
 
 	EsObject* a = opt_array_new ();
 
-	for (int i = 0; i < countXtags (); i++)
+	for (unsigned int i = 0; i < countXtags (); i++)
 	{
 		if (!isTagExtraBitMarked (tag, i))
 			continue;
 
-		langType lang = getXtagOwner (i);
+		langType lang = getXtagLanguage (i);
 		const char *lang_name = (lang == LANG_IGNORE)
 			? NULL
 			: getLanguageName (lang);
@@ -1848,9 +1875,9 @@ static EsObject* getFieldValueForRoles (const tagEntryInfo *tag, const fieldDefi
 static EsObject* getFieldValueForLineCommon (const tagEntryInfo *tag, const fieldDefinition *fdef)
 {
 	if (fdef->ftype == FIELD_END_LINE)
-		return ((int)tag->extensionFields.endLine == 0)
+		return ((int)getTagEndLine (tag) == 0)
 			? es_nil
-			: es_integer_new ((int)tag->extensionFields.endLine);
+			: es_integer_new ((int)getTagEndLine (tag));
 	else
 		return ((int)tag->lineNumber == 0)
 			? es_nil
@@ -1864,7 +1891,7 @@ static EsObject* checkFieldValueForLineCommon (const fieldDefinition *fdef, cons
 
 static EsObject* setFieldValueForLineCommon (tagEntryInfo *tag, const fieldDefinition *fdef, const EsObject *obj)
 {
-	int l;
+	unsigned int l;
 	if (es_object_get_type (obj) == OPT_TYPE_MATCHLOC)
 	{
 		matchLoc *loc = es_pointer_get (obj);
@@ -1872,27 +1899,25 @@ static EsObject* setFieldValueForLineCommon (tagEntryInfo *tag, const fieldDefin
 	}
 	else if (es_integer_p (obj))
 	{
-		l = es_integer_get (obj);
-		if (l < 1)
+		int l0 = es_integer_get (obj);
+		if (l0 < 1)
 			return OPT_ERR_RANGECHECK;
 
+		l = (unsigned int)l0;
 		/* If the new line number is too large,
-		   we cannot fill tag->filePosition wit
+		   we cannot fill tag->filePosition with
 		   getInputFilePositionForLine(); */
 		if (fdef->ftype == FIELD_LINE_NUMBER
-			&& l < getInputLineNumber())
+			&& l > getInputLineNumber())
 			return OPT_ERR_RANGECHECK;
 	}
 	else
 		return OPT_ERR_TYPECHECK;
 
 	if (fdef->ftype == FIELD_END_LINE)
-		tag->extensionFields.endLine = l;
+		setTagEndLine(tag, (unsigned long)l);
 	else
-	{
-		tag->lineNumber = l;
-		tag->filePosition = getInputFilePositionForLine (l);
-	}
+		updateTagLine (tag, l, getInputFilePositionForLine (l));
 
 	return es_false;
 }
@@ -1919,3 +1944,82 @@ static EsObject* setFieldValueForInherits (tagEntryInfo *tag, const fieldDefinit
 
 	return es_false;
 }
+
+extern EsObject* getFieldValueGeneric (const tagEntryInfo *tag, const fieldDefinition *fdef)
+{
+	const char *value = getParserFieldValueForType(tag, fdef->ftype);
+	unsigned int dt = fdef->dataType;
+
+	if (dt & FIELDTYPE_STRING)
+	{
+		if (value == NULL)
+			return es_nil;
+		return (dt & FIELDTYPE_BOOL && value[0] == '\0')
+			? es_false
+			: opt_string_new_from_cstr (value);
+	}
+	else if (dt & FIELDTYPE_INTEGER)
+	{
+		long tmp;
+
+		if (value == NULL)
+			return es_nil;
+		else if (value[0] == '\0')
+			tmp = 0;
+		else if (!strToLong (value, 10, &tmp))
+			tmp = 1;
+
+		return es_integer_new ((int)tmp);
+	}
+	else if (dt & FIELDTYPE_BOOL)
+		return value? es_true: es_nil;
+	else
+	{
+		AssertNotReached ();
+		return es_nil;
+	}
+}
+
+extern EsObject* setFieldValueGeneric (tagEntryInfo *tag, const fieldDefinition *fdef, const EsObject *obj)
+{
+	unsigned int dt = fdef->dataType;
+	const char * val;
+	char buf[1 /* [+-] */ + 20 + 1 /* for \0 */];
+
+	if (dt & FIELDTYPE_STRING)
+	{
+		if (es_object_get_type (obj) == OPT_TYPE_STRING)
+			val = opt_string_get_cstr (obj);
+		else if ((dt & FIELDTYPE_BOOL) && es_object_equal (es_false, obj))
+			val = "";
+		else
+			return OPT_ERR_TYPECHECK;
+	}
+	else if (dt & FIELDTYPE_INTEGER)
+	{
+		int tmp = es_integer_get (obj);
+		/* 2^64 => "18446744073709551616" */
+		snprintf(buf, sizeof(buf), "%d", tmp);
+		val = buf;
+	}
+	else if (dt & FIELDTYPE_BOOL)
+	{
+		if (es_boolean_get (obj))
+			val = "";
+		else
+		{
+			if (doesFieldHaveValue(fdef->ftype, tag))
+				return OPTSCRIPT_ERR_FIELDRESET;
+			val = NULL;
+		}
+	}
+	else
+	{
+		val = "";
+		AssertNotReached ();
+	}
+
+	if (val)
+		attachParserField (tag, fdef->ftype, val);
+	return es_false;
+}
