@@ -18,7 +18,7 @@
 #include "debug.h"
 #include "entry.h"
 #include "htable.h"
-#include "cpreprocessor.h"
+#include "x-cpreprocessor.h"
 #include "kind.h"
 #include "options.h"
 #include "read.h"
@@ -39,6 +39,25 @@
 /*
 *   DATA DECLARATIONS
 */
+enum eCppCharacters {
+	/* white space characters */
+	SPACE         = ' ',
+	NEWLINE       = '\n',
+	CRETURN       = '\r',
+	FORMFEED      = '\f',
+	TAB           = '\t',
+	VTAB          = '\v',
+
+	/* some hard to read characters */
+	DOUBLE_QUOTE  = '"',
+	SINGLE_QUOTE  = '\'',
+	BACKSLASH     = '\\',
+
+	/* symbolic representations, above 0xFF not to conflict with any byte */
+	STRING_SYMBOL = CPP_STRING_SYMBOL,
+	CHAR_SYMBOL   = CPP_CHAR_SYMBOL
+};
+
 typedef enum { COMMENT_NONE, COMMENT_C, COMMENT_CPLUS, COMMENT_D } Comment;
 
 enum eCppLimits {
@@ -87,14 +106,24 @@ enum eState {
 
 /*  Defines the current state of the pre-processor.
  */
+typedef struct sUngetBuffer {
+	unsigned char *buffer;	  /* memory buffer for unget characters */
+	int size;		/* the current unget buffer size */
+	unsigned char *pointer;	  /* the current unget char: points in the
+						   middle of the buffer */
+	int dataSize;		/* the number of valid unget characters
+						   in the buffer */
+	unsigned long lineNumber;
+	MIOPos filePosition;
+	cppMacroInfo *macro;
+} ungetBuffer;
+
 typedef struct sCppState {
 	langType lang;
 	langType clientLang;
 
-	int * ungetBuffer;       /* memory buffer for unget characters */
-	int ungetBufferSize;      /* the current unget buffer size */
-	int * ungetPointer;      /* the current unget char: points in the middle of the buffer */
-	int ungetDataSize;        /* the number of valid unget characters in the buffer */
+	ungetBuffer  *ungetBuffer;
+	ptrArray *ungetBufferStack;
 
 	/* the contents of the last SYMBOL_CHAR or SYMBOL_STRING */
 	vString * charOrStringContents;
@@ -135,6 +164,33 @@ typedef struct sCppState {
 
 } cppState;
 
+#define CPP_MACRO_REPLACEMENT_FLAG_VARARGS 1
+#define CPP_MACRO_REPLACEMENT_FLAG_STRINGIFY 2
+
+struct sCppMacroReplacementPartInfo {
+	int parameterIndex; /* -1 if this part is a constant */
+	int flags;
+	vString * constant; /* not NULL only if parameterIndex != -1 */
+	struct sCppMacroReplacementPartInfo * next;
+};
+
+struct sCppMacroArg {
+	const char *str;
+	unsigned long lineNumber;
+	MIOPos filePosition;
+	bool free_str;
+};
+
+struct sCppMacroTokens {
+	ptrArray *tarray;
+	cppMacroInfo *macro;
+};
+
+typedef struct sCppMacroToken {
+	const char *str;
+	unsigned long lineNumber;
+	MIOPos filePosition;
+} cppMacroToken;
 
 typedef enum {
 	CPREPRO_MACRO_KIND_UNDEF_ROLE,
@@ -158,10 +214,6 @@ static roleDefinition CPREPROHeaderRoles [] = {
 };
 
 
-typedef enum {
-	CPREPRO_MACRO, CPREPRO_HEADER, CPREPRO_PARAM,
-} cPreProkind;
-
 static kindDefinition CPreProKinds [] = {
 	{ true,  'd', "macro",      "macro definitions",
 	  .referenceOnly = false, ATTACH_ROLES(CPREPROMacroRoles)},
@@ -200,12 +252,12 @@ static int externalParserBlockNestLevel;
  */
 static bool BraceFormat = false;
 
-void cppPushExternalParserBlock(void)
+extern void cppPushExternalParserBlock(void)
 {
 	externalParserBlockNestLevel++;
 }
 
-void cppPopExternalParserBlock(void)
+extern void cppPopExternalParserBlock(void)
 {
 	externalParserBlockNestLevel--;
 }
@@ -215,9 +267,7 @@ static cppState Cpp = {
 	.lang = LANG_IGNORE,
 	.clientLang = LANG_IGNORE,
 	.ungetBuffer = NULL,
-	.ungetBufferSize = 0,
-	.ungetPointer = NULL,
-	.ungetDataSize = 0,
+	.ungetBufferStack = NULL,
 	.charOrStringContents = NULL,
 	.resolveRequired = false,
 	.hasAtLiteralStrings = false,
@@ -257,6 +307,8 @@ static cppState Cpp = {
 static hashTable *makeMacroTable (void);
 static cppMacroInfo * saveMacro(hashTable *table, const char * macro);
 
+static void cppMacroTokensDelete (cppMacroTokens *tokens);
+
 /*
 *   FUNCTION DEFINITIONS
 */
@@ -271,6 +323,31 @@ extern unsigned int cppGetDirectiveNestLevel (void)
 	return Cpp.directive.nestLevel;
 }
 
+static ungetBuffer *ungetBufferNew  (unsigned long lineNumber,
+									 MIOPos filePosition,
+									 cppMacroInfo *macro)
+{
+	ungetBuffer *ub = xCalloc (1, ungetBuffer);
+	ub->lineNumber = lineNumber;
+	ub->filePosition = filePosition;
+	ub->macro = macro;
+	return ub;
+}
+
+static void ungetBufferDelete (ungetBuffer *ub)
+{
+	if (!ub)
+		return;
+
+	if (ub->buffer)
+	{
+		eFree (ub->buffer);
+		ub->buffer = NULL;
+	}
+
+	eFree (ub);
+}
+
 static void cppInitCommon(langType clientLang,
 		     const bool state, const bool hasAtLiteralStrings,
 		     const bool hasCxxRawLiteralStrings,
@@ -299,7 +376,7 @@ static void cppInitCommon(langType clientLang,
 
 	Cpp.clientLang = clientLang;
 	Cpp.ungetBuffer = NULL;
-	Cpp.ungetPointer = NULL;
+	Cpp.ungetBufferStack = ptrArrayNew ((ptrArrayDeleteFunc)ungetBufferDelete);
 
 	CXX_DEBUG_ASSERT(!Cpp.charOrStringContents,"This string should be null when CPP is not initialized");
 	Cpp.charOrStringContents = vStringNew();
@@ -418,11 +495,10 @@ extern void cppTerminate (void)
 		Cpp.directive.name = NULL;
 	}
 
-	if(Cpp.ungetBuffer)
-	{
-		eFree(Cpp.ungetBuffer);
-		Cpp.ungetBuffer = NULL;
-	}
+	ungetBufferDelete (Cpp.ungetBuffer); /* NULL is acceptable */
+	Cpp.ungetBuffer = NULL;
+	ptrArrayDelete (Cpp.ungetBufferStack);
+	Cpp.ungetBufferStack = NULL;
 
 	if(Cpp.charOrStringContents)
 	{
@@ -459,104 +535,177 @@ extern void cppEndStatement (void)
 */
 
 /*  This puts a character back into the input queue for the input File. */
-extern void cppUngetc (const int c)
+static void ungetBufferUngetc (ungetBuffer *ungetBuffer, const int c, vString *charOrStringContents)
 {
-	if(!Cpp.ungetPointer)
+	if (c == STRING_SYMBOL || c == CHAR_SYMBOL)
+	{
+		Assert(charOrStringContents != NULL);
+		cppUngetc(c == STRING_SYMBOL ? '"' : '\'');
+		cppUngetString(vStringValue(charOrStringContents), vStringLength(charOrStringContents));
+		cppUngetc(c == STRING_SYMBOL ? '"' : '\'');
+		vStringClear(charOrStringContents);
+		return;
+	}
+	else if (c == EOF)
+	{
+		return;
+	}
+
+	Assert((unsigned int)c <= 0xff);
+	unsigned char u = (unsigned char)c;
+
+	if(!ungetBuffer->pointer)
 	{
 		// no unget data
-		if(!Cpp.ungetBuffer)
+		if(!ungetBuffer->buffer)
 		{
-			Cpp.ungetBuffer = (int *)eMalloc(8 * sizeof(int));
-			Cpp.ungetBufferSize = 8;
+			ungetBuffer->buffer = eMalloc(8 * sizeof(*(ungetBuffer->buffer)));
+			ungetBuffer->size = 8;
 		}
-		Assert(Cpp.ungetBufferSize > 0);
-		Cpp.ungetPointer = Cpp.ungetBuffer + Cpp.ungetBufferSize - 1;
-		*(Cpp.ungetPointer) = c;
-		Cpp.ungetDataSize = 1;
+		Assert(ungetBuffer->size > 0);
+		ungetBuffer->pointer = ungetBuffer->buffer + ungetBuffer->size - 1;
+		*(ungetBuffer->pointer) = u;
+		ungetBuffer->dataSize = 1;
 		return;
 	}
 
 	// Already have some unget data in the buffer. Must prepend.
-	Assert(Cpp.ungetBuffer);
-	Assert(Cpp.ungetBufferSize > 0);
-	Assert(Cpp.ungetDataSize > 0);
-	Assert(Cpp.ungetPointer >= Cpp.ungetBuffer);
+	Assert(ungetBuffer->buffer);
+	Assert(ungetBuffer->size > 0);
+	Assert(ungetBuffer->dataSize > 0);
+	Assert(ungetBuffer->pointer >= ungetBuffer->buffer);
 
-	if(Cpp.ungetPointer == Cpp.ungetBuffer)
+	if(ungetBuffer->pointer == ungetBuffer->buffer)
 	{
-		Cpp.ungetBufferSize += 8;
-		int * tmp = (int *)eMalloc(Cpp.ungetBufferSize * sizeof(int));
-		memcpy(tmp+8,Cpp.ungetPointer,Cpp.ungetDataSize * sizeof(int));
-		eFree(Cpp.ungetBuffer);
-		Cpp.ungetBuffer = tmp;
-		Cpp.ungetPointer = tmp + 7;
+		ungetBuffer->size += 8;
+		unsigned char * tmp = eMalloc(ungetBuffer->size * sizeof(*(ungetBuffer->buffer)));
+		memcpy(tmp+8,ungetBuffer->pointer,ungetBuffer->dataSize * sizeof(*(ungetBuffer->buffer)));
+		eFree(ungetBuffer->buffer);
+		ungetBuffer->buffer = tmp;
+		ungetBuffer->pointer = tmp + 7;
 	} else {
-		Cpp.ungetPointer--;
+		ungetBuffer->pointer--;
 	}
 
-	*(Cpp.ungetPointer) = c;
-	Cpp.ungetDataSize++;
+	*(ungetBuffer->pointer) = u;
+	ungetBuffer->dataSize++;
 }
 
-int cppUngetBufferSize()
+static int ungetBufferSize (ungetBuffer *ungetBuffer)
 {
-	return Cpp.ungetBufferSize;
+	return ungetBuffer->size;
 }
 
-/*  This puts an entire string back into the input queue for the input File. */
-void cppUngetString(const char * string,int len)
+static void ungetBufferUngetString(ungetBuffer *ungetBuffer, const char * string, int len)
 {
 	if(!string)
 		return;
 	if(len < 1)
 		return;
 
-	if(!Cpp.ungetPointer)
+	if(!ungetBuffer->pointer)
 	{
 		// no unget data
-		if(!Cpp.ungetBuffer)
+		if(!ungetBuffer->buffer)
 		{
-			Cpp.ungetBufferSize = 8 + len;
-			Cpp.ungetBuffer = (int *)eMalloc(Cpp.ungetBufferSize * sizeof(int));
-		} else if(Cpp.ungetBufferSize < len)
+			ungetBuffer->size = 8 + len;
+			ungetBuffer->buffer = eMalloc(ungetBuffer->size * sizeof(*(ungetBuffer->buffer)));
+		} else if(ungetBuffer->size < len)
 		{
-			Cpp.ungetBufferSize = 8 + len;
-			Cpp.ungetBuffer = (int *)eRealloc(Cpp.ungetBuffer,Cpp.ungetBufferSize * sizeof(int));
+			ungetBuffer->size = 8 + len;
+			ungetBuffer->buffer = eRealloc(ungetBuffer->buffer,
+										   ungetBuffer->size * sizeof(*(ungetBuffer->buffer)));
 		}
-		Cpp.ungetPointer = Cpp.ungetBuffer + Cpp.ungetBufferSize - len;
+		ungetBuffer->pointer = ungetBuffer->buffer + ungetBuffer->size - len;
 	} else {
 		// Already have some unget data in the buffer. Must prepend.
-		Assert(Cpp.ungetBuffer);
-		Assert(Cpp.ungetBufferSize > 0);
-		Assert(Cpp.ungetDataSize > 0);
-		Assert(Cpp.ungetPointer >= Cpp.ungetBuffer);
+		Assert(ungetBuffer->buffer);
+		Assert(ungetBuffer->size > 0);
+		Assert(ungetBuffer->dataSize > 0);
+		Assert(ungetBuffer->pointer >= ungetBuffer->buffer);
 
-		if(Cpp.ungetBufferSize < (Cpp.ungetDataSize + len))
+		if(ungetBuffer->size < (ungetBuffer->dataSize + len))
 		{
-			Cpp.ungetBufferSize = 8 + len + Cpp.ungetDataSize;
-			int * tmp = (int *)eMalloc(Cpp.ungetBufferSize * sizeof(int));
-			memcpy(tmp + 8 + len,Cpp.ungetPointer,Cpp.ungetDataSize * sizeof(int));
-			eFree(Cpp.ungetBuffer);
-			Cpp.ungetBuffer = tmp;
-			Cpp.ungetPointer = tmp + 8;
+			ungetBuffer->size = 8 + len + ungetBuffer->dataSize;
+			unsigned char * tmp = eMalloc(ungetBuffer->size * sizeof(*(ungetBuffer->buffer)));
+			memcpy(tmp + 8 + len,ungetBuffer->pointer,ungetBuffer->dataSize * sizeof(*(ungetBuffer->buffer)));
+			eFree(ungetBuffer->buffer);
+			ungetBuffer->buffer = tmp;
+			ungetBuffer->pointer = tmp + 8;
 		} else {
-			Cpp.ungetPointer -= len;
-			Assert(Cpp.ungetPointer >= Cpp.ungetBuffer);
+			ungetBuffer->pointer -= len;
+			Assert(ungetBuffer->pointer >= ungetBuffer->buffer);
 		}
 	}
 
-	int * p = Cpp.ungetPointer;
+	unsigned char* p = ungetBuffer->pointer;
 	const char * s = string;
 	const char * e = string + len;
 
 	while(s < e)
-		*p++ = *s++;
+		*p++ = (unsigned char)(*s++);
 
-	Cpp.ungetDataSize += len;
+	ungetBuffer->dataSize += len;
 }
 
-extern void cppUngetStringBuiltByMacro(const char * string,int len, cppMacroInfo *macro)
+static int ungetBufferGetcFromUngetBuffer (ungetBuffer *ungetBuffer)
 {
+	if(ungetBuffer->pointer)
+	{
+		Assert(ungetBuffer->buffer);
+		Assert(ungetBuffer->size > 0);
+		Assert(ungetBuffer->dataSize > 0);
+
+		int c = *(ungetBuffer->pointer);
+		ungetBuffer->dataSize--;
+		if(ungetBuffer->dataSize > 0)
+			ungetBuffer->pointer++;
+		else
+			ungetBuffer->pointer = NULL;
+		return c;
+	}
+	return EOF;
+}
+
+extern void cppUngetc (const int c)
+{
+	if (Cpp.ungetBuffer == NULL)
+		Cpp.ungetBuffer = ungetBufferNew (getInputLineNumber(),
+										  getInputFilePosition(),
+										  NULL);
+	ungetBufferUngetc (Cpp.ungetBuffer , c, Cpp.charOrStringContents);
+}
+
+extern int cppUngetBufferSize(void)
+{
+	if (Cpp.ungetBuffer == NULL)
+		return 0;
+
+	int size = ungetBufferSize (Cpp.ungetBuffer);
+	for (size_t i = 0; i < ptrArrayCount (Cpp.ungetBufferStack); i++)
+	{
+		ungetBuffer *ub = ptrArrayItem (Cpp.ungetBufferStack, i);
+		i += ungetBufferSize (ub);
+	}
+	return size;
+}
+
+/*  This puts an entire string back into the input queue for the input File. */
+extern void cppUngetString(const char * string, int len)
+{
+	if (Cpp.ungetBuffer == NULL)
+		Cpp.ungetBuffer = ungetBufferNew (getInputLineNumber(),
+										  getInputFilePosition(),
+										  NULL);
+	ungetBufferUngetString (Cpp.ungetBuffer, string, len);
+}
+
+extern void cppUngetMacroTokens (cppMacroTokens *tokens)
+{
+	Assert (tokens);
+
+	cppMacroInfo *macro = tokens->macro;
+
 	if (macro->useCount == 0)
 	{
 		cppMacroInfo *m = Cpp.macroInUse;
@@ -568,31 +717,72 @@ extern void cppUngetStringBuiltByMacro(const char * string,int len, cppMacroInfo
 	CXX_DEBUG_PRINT("Macro <%p> increment useCount: %d->%d", macro,
 					(macro->useCount - 1), macro->useCount);
 
-	cppUngetString (string, len);
+	ptrArray *a = tokens->tarray;
+	if (ptrArrayIsEmpty (a))
+	{
+		cppMacroTokensDelete (tokens);
+		return;
+	}
+
+	if (Cpp.ungetBuffer)
+	{
+		ptrArrayAdd(Cpp.ungetBufferStack, Cpp.ungetBuffer);
+		Cpp.ungetBuffer = NULL;
+	}
+
+	for(size_t i = ptrArrayCount (a); i > 0; i--)
+	{
+		cppMacroToken *t = ptrArrayItem (a, i - 1);
+		ungetBuffer *ub = ungetBufferNew(t->lineNumber, t->filePosition,
+										 macro);
+		ungetBufferUngetString (ub, t->str, strlen (t->str));
+		ptrArrayAdd(Cpp.ungetBufferStack, ub);
+	}
+
+	if (!ptrArrayIsEmpty (Cpp.ungetBufferStack))
+		Cpp.ungetBuffer = ptrArrayRemoveLast (Cpp.ungetBufferStack);
+	cppMacroTokensDelete (tokens);
 }
 
 static int cppGetcFromUngetBufferOrFile(void)
 {
-	if(Cpp.ungetPointer)
+ retry:
+	if (Cpp.ungetBuffer)
 	{
-		Assert(Cpp.ungetBuffer);
-		Assert(Cpp.ungetBufferSize > 0);
-		Assert(Cpp.ungetDataSize > 0);
+		int c = ungetBufferGetcFromUngetBuffer (Cpp.ungetBuffer);
+		if (c != EOF)
+			return c;
 
-		int c = *(Cpp.ungetPointer);
-		Cpp.ungetDataSize--;
-		if(Cpp.ungetDataSize > 0)
-			Cpp.ungetPointer++;
-		else
-			Cpp.ungetPointer = NULL;
-		return c;
+		ungetBufferDelete (Cpp.ungetBuffer);
+		Cpp.ungetBuffer = NULL;
+		if (!ptrArrayIsEmpty (Cpp.ungetBufferStack))
+		{
+			Cpp.ungetBuffer = ptrArrayRemoveLast (Cpp.ungetBufferStack);
+			goto retry;
+		}
 	}
 
+	/* Or */
+
 	if (Cpp.macroInUse)
 		cppClearMacroInUse (&Cpp.macroInUse);
 	return getcFromInputFile();
 }
 
+extern unsigned long cppGetInputLineNumber (void)
+{
+	if (Cpp.ungetBuffer)
+		return Cpp.ungetBuffer->lineNumber;
+	return getInputLineNumber();
+}
+
+extern MIOPos cppGetInputFilePosition (void)
+{
+	if (Cpp.ungetBuffer)
+		return Cpp.ungetBuffer->filePosition;
+	return getInputFilePosition();
+}
+
 
 /*  Reads a directive, whose first character is given by "c", into "name".
  */
@@ -871,7 +1061,7 @@ static void makeIncludeTag (const  char *const name, bool systemHeader)
 	}
 }
 
-static void makeParamTag (vString *name, short nth, bool placeholder)
+static int makeParamTag (vString *name, short nth, bool placeholder)
 {
 	bool standing_alone = doesCPreProRunAsStandaloneParser(CPREPRO_MACRO);
 
@@ -879,26 +1069,29 @@ static void makeParamTag (vString *name, short nth, bool placeholder)
 
 	if (standing_alone)
 		pushLanguage (Cpp.lang);
-	int r = makeSimpleTag (name, Cpp.macroParamKindIndex);
+
+	tagEntryInfo e;
+	initTagEntry (&e, vStringValue (name), Cpp.macroParamKindIndex);
+	updateTagLine (&e, cppGetInputLineNumber (), cppGetInputFilePosition ());
+	e.extensionFields.nth = nth;
+	if (placeholder)
+		markTagAsPlaceholder (&e, placeholder);
+	int r = makeTagEntry(&e);
+
 	if (standing_alone)
 		popLanguage ();
 
-	tagEntryInfo *e = getEntryInCorkQueue (r);
-	if (e)
-	{
-		e->extensionFields.nth = nth;
-		if (placeholder)
-			markTagAsPlaceholder (e, placeholder);
-	}
+	return r;
 }
 
-static void regenreateSignatureFromParameters (vString * buffer, int from, int to)
+static void makeSignatureStringFromParameters (vString * buffer, intArray *parameters)
 {
 	vStringPut(buffer, '(');
-	for (int pindex = from; pindex < to; pindex++)
+	for (size_t i = 0; i < intArrayCount (parameters); i++)
 	{
+		int pindex = intArrayItem (parameters, i);
 		tagEntryInfo *e = getEntryInCorkQueue (pindex);
-		if (e && !isTagExtra (e))
+		if (e)
 		{
 			vStringCatS (buffer, e->name);
 			vStringPut (buffer, ',');
@@ -932,13 +1125,14 @@ static int directiveDefine (const int c, bool undef)
 		readIdentifier (c, Cpp.directive.name);
 		if (! isIgnore ())
 		{
-			unsigned long 	lineNumber = getInputLineNumber ();
-			MIOPos filePosition = getInputFilePosition ();
+			unsigned long 	lineNumber = cppGetInputLineNumber ();
+			MIOPos filePosition = cppGetInputFilePosition ();
 			int p = cppGetcFromUngetBufferOrFile ();
 			short nth = 0;
 
 			if (p == '(')
 			{
+				intArray *params = intArrayNew ();
 				vString *param = vStringNew ();
 				int param_start = (int)countEntryInCorkQueue();
 				do {
@@ -953,7 +1147,29 @@ static int directiveDefine (const int c, bool undef)
 
 					if (vStringLength (param) > 0)
 					{
-						makeParamTag (param, nth++, vStringChar(param, 0) == '.');
+						bool gnuext_placeholder = false;
+						if (vStringLength (param) > 3
+							&& strcmp(vStringValue (param) + vStringLength (param) - 3,
+									  "...")  == 0)
+						{
+							/* args... in GNU cpp extension
+							 *
+							 * #define debug(format, args...) fprintf (stderr, format, args)
+							 *
+							 * In this case, args should be tagged. However the signature field
+							 * for debug must be "(format,args...)".
+							 */
+							vString *nodots = vStringNewNInit (vStringValue (param),
+															   vStringLength (param) - 3);
+							makeParamTag (nodots, nth, false);
+							vStringDelete (nodots);
+							gnuext_placeholder = true;
+						}
+
+						int r = makeParamTag (param, nth++,
+											  vStringChar(param, 0) == '.'
+											  || gnuext_placeholder);
+						intArrayAdd (params, r);
 						vStringClear (param);
 					}
 					if (p == '\\')
@@ -965,18 +1181,18 @@ static int directiveDefine (const int c, bool undef)
 				if (p == ')')
 				{
 					vString *signature = vStringNew ();
-					regenreateSignatureFromParameters (signature, param_start, param_end);
+					makeSignatureStringFromParameters (signature, params);
 					r = makeDefineTag (vStringValue (Cpp.directive.name), vStringValue (signature), undef);
 					vStringDelete (signature);
 				}
 				else
 					r = makeDefineTag (vStringValue (Cpp.directive.name), NULL, undef);
+				intArrayDelete (params);
 
 				tagEntryInfo *e = getEntryInCorkQueue (r);
 				if (e)
 				{
-					e->lineNumber = lineNumber;
-					e->filePosition = filePosition;
+					updateTagLine (e, lineNumber, filePosition);
 					patchScopeFieldOfParameters (param_start, param_end, r);
 				}
 			}
@@ -1082,7 +1298,7 @@ static bool directiveIf (const int c, enum eIfSubstate if_substate)
 	{
 		conditionalInfo *ifdef = currentConditional ();
 		ifdef->asmArea.ifSubstate = if_substate;
-		ifdef->asmArea.line = getInputLineNumber();
+		ifdef->asmArea.line = cppGetInputLineNumber ();
 	}
 
 	Cpp.directive.state = DRCTV_NONE;
@@ -1120,7 +1336,7 @@ static void promiseOrPrepareAsm (conditionalInfo *ifdef, enum eIfSubstate curren
 			&& (currentState == IF_ENDIF)))
 	{
 		unsigned long start = ifdef->asmArea.line + 1;
-		unsigned long end = getInputLineNumber ();
+		unsigned long end = cppGetInputLineNumber ();
 
 		if (start < end)
 			makePromise ("Asm", start, 0, end, 0, start);
@@ -1134,7 +1350,7 @@ static void promiseOrPrepareAsm (conditionalInfo *ifdef, enum eIfSubstate curren
 		else if (currentState == IF_ELSE)
 		{
 			ifdef->asmArea.ifSubstate = IF_ELSE;
-			ifdef->asmArea.line = getInputLineNumber ();
+			ifdef->asmArea.line = cppGetInputLineNumber ();
 		}
 	}
 }
@@ -1313,7 +1529,7 @@ static int skipOverDComment (void)
 	return c;
 }
 
-const vString * cppGetLastCharOrStringContents (void)
+extern const vString * cppGetLastCharOrStringContents (void)
 {
 	CXX_DEBUG_ASSERT(Cpp.charOrStringContents,"Shouldn't be called when CPP is not initialized");
 	return Cpp.charOrStringContents;
@@ -1349,7 +1565,7 @@ static int skipToEndOfString (bool ignoreBackslash)
 		else if (c == DOUBLE_QUOTE)
 			break;
 		else
-			vStringPutWithLimit (Cpp.charOrStringContents, c, 1024);
+			(void)vStringPutWithLimit (Cpp.charOrStringContents, c, 1024);
 	}
 	return STRING_SYMBOL;  /* symbolic representation of string */
 }
@@ -1407,7 +1623,7 @@ static int skipToEndOfCxxRawLiteralString (void)
  *  special character to symbolically represent a generic character.
  *  Also detects Vera numbers that include a base specifier (ie. 'b1010).
  */
-static int skipToEndOfChar ()
+static int skipToEndOfChar (void)
 {
 	int c;
 	int count = 0, veraBase = '\0';
@@ -1445,7 +1661,7 @@ static int skipToEndOfChar ()
 			if (count == 1  &&  strchr ("DHOB", toupper (c)) != NULL)
 			{
 				veraBase = c;
-				vStringPutWithLimit (Cpp.charOrStringContents, c, 10);
+				(void)vStringPutWithLimit (Cpp.charOrStringContents, c, 10);
 			}
 			else if (veraBase != '\0'  &&  ! isalnum (c))
 			{
@@ -1453,10 +1669,10 @@ static int skipToEndOfChar ()
 				break;
 			}
 			else
-				vStringPutWithLimit (Cpp.charOrStringContents, c, 10);
+				(void)vStringPutWithLimit (Cpp.charOrStringContents, c, 10);
 		}
 		else
-			vStringPutWithLimit (Cpp.charOrStringContents, c, 10);
+			(void)vStringPutWithLimit (Cpp.charOrStringContents, c, 10);
 	}
 	return CHAR_SYMBOL;  /* symbolic representation of character */
 }
@@ -1464,12 +1680,12 @@ static int skipToEndOfChar ()
 static void attachFields (int macroCorkIndex, unsigned long endLine, const char *macrodef)
 {
 	tagEntryInfo *tag = getEntryInCorkQueue (macroCorkIndex);
-	if (!tag)
-		return;
-
-	tag->extensionFields.endLine = endLine;
-	if (macrodef)
-		attachParserFieldToCorkEntry (macroCorkIndex, Cpp.macrodefFieldIndex, macrodef);
+	if (tag)
+	{
+		setTagEndLine (tag, endLine);
+		if (macrodef)
+			attachParserField (tag, Cpp.macrodefFieldIndex, macrodef);
+	}
 }
 
 static vString * conditionMayFlush (vString* condition, bool del)
@@ -1489,7 +1705,11 @@ static vString * conditionMayFlush (vString* condition, bool del)
 		if (standing_alone)
 			pushLanguage (Cpp.lang);
 
-		makeSimpleRefTag (condition, Cpp.defineMacroKindIndex, Cpp.macroConditionRoleIndex);
+		tagEntryInfo e;
+		initRefTagEntry (&e, vStringValue (condition),
+						 Cpp.defineMacroKindIndex, Cpp.macroConditionRoleIndex);
+		updateTagLine (&e, cppGetInputLineNumber (), cppGetInputFilePosition ());
+		makeTagEntry (&e);
 
 		if (standing_alone)
 			popLanguage ();
@@ -1515,6 +1735,29 @@ static void conditionMayPut (vString *condition, int c)
 		vStringPut(condition, c);
 }
 
+extern void cppVStringPut (vString* string, const int c)
+{
+	if (c <= 0xff)
+		vStringPut (string, c);
+	else
+	{
+		char marker = '"';
+		switch (c)
+		{
+			case CHAR_SYMBOL:
+				marker = '\'';
+				/* Fall through */
+			case STRING_SYMBOL:
+				vStringPut (string, marker);
+				vStringCat (string, cppGetLastCharOrStringContents ());
+				vStringPut (string, marker);
+				break;
+			default:
+				AssertNotReached();
+		}
+	}
+}
+
 /*  This function returns the next character, stripping out comments,
  *  C pre-processor directives, and the contents of single and double
  *  quoted strings. In short, strip anything which places a burden upon
@@ -1542,7 +1785,7 @@ process:
 				if (macroCorkIndex != CORK_NIL)
 				{
 					attachFields (macroCorkIndex,
-								  getInputLineNumber(),
+								  cppGetInputLineNumber(),
 								  macrodef? vStringValue (macrodef): NULL);
 					macroCorkIndex = CORK_NIL;
 				}
@@ -1566,7 +1809,7 @@ process:
 					if (macroCorkIndex != CORK_NIL)
 					{
 						attachFields (macroCorkIndex,
-									  getInputLineNumber(),
+									  cppGetInputLineNumber(),
 									  macrodef? vStringValue (macrodef): NULL);
 						macroCorkIndex = CORK_NIL;
 					}
@@ -1912,9 +2155,9 @@ process:
 	if (condition)
 		vStringDelete (condition);
 
-	DebugStatement ( debugPutc (DEBUG_CPP, c); )
+	DebugStatement ( cppDebugPutc (DEBUG_CPP, c); )
 	DebugStatement ( if (c == NEWLINE)
-				debugPrintf (DEBUG_CPP, "%6ld: ", getInputLineNumber () + 1); )
+				debugPrintf (DEBUG_CPP, "%6ld: ", cppGetInputLineNumber () + 1); )
 
 	return c;
 }
@@ -1967,7 +2210,7 @@ static bool buildMacroInfoFromTagEntry (int corkIndex,
 	return true;
 }
 
-extern cppMacroInfo * cppFindMacroFromSymtab (const char *const name)
+static cppMacroInfo * cppFindMacroFromSymtab (const char *const name)
 {
 	cppMacroInfo *info = NULL;
 	foreachEntriesInScope (CORK_NIL, name, buildMacroInfoFromTagEntry, &info);
@@ -2001,20 +2244,87 @@ extern cppMacroInfo * cppFindMacro (const char *const name)
 	return NULL;
 }
 
-extern vString * cppBuildMacroReplacement(
-		const cppMacroInfo * macro,
-		const char ** parameters, /* may be NULL */
-		int parameterCount
-	)
+extern cppMacroArg *cppMacroArgNew (const char *str, bool free_str_when_deleting,
+									unsigned long lineNumber, MIOPos filePosition)
+{
+	cppMacroArg *a = xMalloc (1, cppMacroArg);
+
+	a->str = str;
+	a->free_str = free_str_when_deleting;
+	a->lineNumber = lineNumber;
+	a->filePosition = filePosition;
+
+	return a;
+}
+
+extern void cppMacroArgDelete (void *macroArg)
+{
+	cppMacroArg *a = (cppMacroArg *)macroArg;
+	if (a->free_str)
+		eFree ((void *)a->str);
+	a->str = NULL;
+	eFree (macroArg);
+}
+
+static cppMacroToken *cppMacroTokenNew(unsigned long lineNumber, MIOPos filePosition)
+{
+	cppMacroToken *t = xMalloc (1, cppMacroToken);
+	t->str = NULL;
+	t->lineNumber = lineNumber;
+	t->filePosition = filePosition;
+	return t;
+}
+
+static void cppMacroTokenDelete (cppMacroToken *t)
+{
+	if (t->str)
+		eFree ((void *)t->str);
+	t->str = NULL;
+	eFree (t);
+}
+
+static cppMacroTokens *cppMacroTokensNew (cppMacroInfo * macro)
+{
+	cppMacroTokens *r = xMalloc (1, cppMacroTokens);
+
+	r->tarray = ptrArrayNew ((ptrArrayDeleteFunc)cppMacroTokenDelete);
+	r->macro = macro;
+
+	return r;
+}
+
+static void cppMacroTokensDelete (cppMacroTokens *tokens)
+{
+	Assert (tokens);
+
+	ptrArrayDelete (tokens->tarray);
+	tokens->tarray = NULL;
+	eFree (tokens);
+}
+
+extern cppMacroTokens *cppExpandMacro (cppMacroInfo * macro,
+									   const ptrArray *args,
+									   unsigned long lineNumber,
+									   MIOPos filePosition)
 {
 	if(!macro)
 		return NULL;
 
+	if(cppUngetBufferSize () >= CPP_MAXIMUM_UNGET_BUFFER_SIZE_FOR_MACRO_REPLACEMENTS)
+	{
+		CXX_DEBUG_PRINT ("Ungetbuffer overflow when processing \"%s\": %d",
+						 macro->name, cppUngetBufferSize());
+		return NULL;
+	}
+
 	if(!macro->replacements)
 		return NULL;
 
-	vString * ret = vStringNew();
+	cppMacroTokens *tokens = cppMacroTokensNew (macro);
+	cppMacroToken *t = cppMacroTokenNew(lineNumber, filePosition);
+	ptrArrayAdd (tokens->tarray, t);
 
+	vString * vstr = vStringNew();
 	cppMacroReplacementPartInfo * r = macro->replacements;
 
 	while(r)
@@ -2022,83 +2332,91 @@ extern vString * cppBuildMacroReplacement(
 		if(r->parameterIndex < 0)
 		{
 			if(r->constant)
-				vStringCat(ret,r->constant);
+			{
+				/* We assume the constant parts of the macro definition are
+				 * at the beginning of the current macro expansion. */
+				t->str = vStringDeleteUnwrap (vstr);
+
+				t = cppMacroTokenNew(lineNumber, filePosition);
+				ptrArrayAdd (tokens->tarray, t);
+
+				vstr = vStringNewCopy (r->constant);
+			}
 		} else {
-			if(parameters && (r->parameterIndex < parameterCount))
+			if(args && (r->parameterIndex < ptrArrayCount (args)))
 			{
 				if(r->flags & CPP_MACRO_REPLACEMENT_FLAG_STRINGIFY)
-					vStringPut(ret,'"');
+					vStringPut(vstr,'"');
+				t->str = vStringDeleteUnwrap (vstr);
 
-				vStringCatS(ret,parameters[r->parameterIndex]);
+				cppMacroArg *a = ptrArrayItem (args, r->parameterIndex);
+				t = cppMacroTokenNew(a->lineNumber, a->filePosition);
+				ptrArrayAdd (tokens->tarray, t);
+
+				vstr = vStringNewInit(a->str);
 				if(r->flags & CPP_MACRO_REPLACEMENT_FLAG_VARARGS)
 				{
 					int idx = r->parameterIndex + 1;
-					while(idx < parameterCount)
+					while(idx < ptrArrayCount (args))
 					{
-						vStringPut(ret,',');
-						vStringCatS(ret,parameters[idx]);
+						vStringPut(vstr,',');
+						t->str = vStringDeleteUnwrap (vstr);
+
+						a = ptrArrayItem (args, idx);
+						t = cppMacroTokenNew(a->lineNumber, a->filePosition);
+						ptrArrayAdd (tokens->tarray, t);
+						vstr = vStringNew();
+
+						vStringCatS(vstr, a->str);
 						idx++;
 					}
 				}
 
 				if(r->flags & CPP_MACRO_REPLACEMENT_FLAG_STRINGIFY)
-					vStringPut(ret,'"');
+					vStringPut(vstr,'"');
 			}
 		}
 
 		r = r->next;
 	}
 
-	return ret;
+	if (t->str == NULL)
+		t->str = vStringDeleteUnwrap (vstr);
+
+	return tokens;
 }
 
-// We stop applying macro replacements if the unget buffer gets too big
-// as it is a sign of recursive macro expansion
-#define CPP_PARSER_MAXIMUM_UNGET_BUFFER_SIZE_FOR_MACRO_REPLACEMENTS 65536
-
-extern void cppBuildMacroReplacementWithPtrArrayAndUngetResult(
-		cppMacroInfo * macro,
-		const ptrArray * args)
+#ifdef DEBUG
+extern
+#else
+static
+#endif
+vString *cppFlattenMacroTokensToNewString (cppMacroTokens *tokens)
 {
-	vString * replacement = NULL;
+	Assert (tokens);
 
-	// Detect other cases of nasty macro expansion that cause
-	// the unget buffer to grow fast (but the token chain to grow slowly)
-	//    -D'p=a' -D'a=p+p'
-	if ((cppUngetBufferSize() < CPP_PARSER_MAXIMUM_UNGET_BUFFER_SIZE_FOR_MACRO_REPLACEMENTS)
-		&& macro->replacements)
+	vString *vstr = vStringNew ();
+
+	for (size_t i = 0; i < ptrArrayCount (tokens->tarray); i++)
 	{
-		int argc = 0;
-		const char ** argv = NULL;
-
-		if (args)
-		{
-			argc = ptrArrayCount (args);
-			argv = (const char **)eMalloc (sizeof(char *) * argc);
-			for (int i = 0; i < argc; i++)
-			{
-				TRACE_PRINT("Arg[%d] for %s<%p>: %s",
-							i, macro->name, macro, ptrArrayItem (args, i));
-				argv[i] = ptrArrayItem (args, i);
-			}
-		}
-
-		replacement = cppBuildMacroReplacement(macro, argv, argc);
-
-		if (argv)
-			eFree ((void *)argv);
+		cppMacroToken *t = ptrArrayItem (tokens->tarray, i);
+		vStringCatS (vstr, t->str);
 	}
 
-	if (replacement)
-	{
-		cppUngetStringBuiltByMacro(vStringValue(replacement), vStringLength(replacement),
-								   macro);
-		TRACE_PRINT("Replacement for %s<%p>: %s", macro->name, macro, vStringValue (replacement));
-		vStringDelete (replacement);
-	}
-	else
-		TRACE_PRINT("Replacement for %s<%p>: ", macro->name, macro);
+	return vstr;
+}
 
+extern vString *cppExpandMacroAsNewString(cppMacroInfo * macro, const ptrArray *args)
+{
+	cppMacroTokens * tokens = cppExpandMacro (macro, args,
+											  cppGetInputLineNumber (),
+											  cppGetInputFilePosition ());
+	if (!tokens)
+		return NULL;
+
+	vString *vstr = cppFlattenMacroTokensToNewString (tokens);
+	cppMacroTokensDelete (tokens);
+	return vstr;
 }
 
 static void saveIgnoreToken(const char * ignoreToken)
@@ -2189,7 +2507,7 @@ static cppMacroInfo * saveMacro(hashTable *table, const char * macro)
 		return NULL;
 	}
 
-	if(!(isalpha(*c) || (*c == '_' || (*c == '$') )))
+	if(!(isalpha((unsigned char) *c) || (*c == '_' || (*c == '$') )))
 	{
 		CXX_DEBUG_LEAVE_TEXT("Macro does not start with an alphanumeric character");
 		return NULL; // must be a sequence of letters and digits
@@ -2197,7 +2515,7 @@ static cppMacroInfo * saveMacro(hashTable *table, const char * macro)
 
 	const char * identifierBegin = c;
 
-	while(*c && (isalnum(*c) || (*c == '_') || (*c == '$') ))
+	while(*c && (isalnum((unsigned char) *c) || (*c == '_') || (*c == '$') ))
 		c++;
 
 	const char * identifierEnd = c;
@@ -2270,6 +2588,7 @@ static cppMacroInfo * saveMacro(hashTable *table, const char * macro)
 		CXX_DEBUG_PRINT("Got %d parameters",iParamCount);
 
 	} else {
+		CXX_DEBUG_PRINT("Macro has no parameter list");
 		info->hasParameterList = false;
 	}
 
@@ -2330,14 +2649,14 @@ static cppMacroInfo * saveMacro(hashTable *table, const char * macro)
 
 		while(*c)
 		{
-			if(isalpha(*c) || (*c == '_'))
+			if(isalpha((unsigned char) *c) || (*c == '_'))
 			{
 				if(c > begin)
 					ADD_CONSTANT_REPLACEMENT(begin,c - begin);
 
 				const char * tokenBegin = c;
 
-				while(*c && (isalnum(*c) || (*c == '_')))
+				while(*c && (isalnum((unsigned char) *c) || (*c == '_')))
 					c++;
 
 				// check if it is a parameter
@@ -2354,13 +2673,25 @@ static cppMacroInfo * saveMacro(hashTable *table, const char * macro)
 
 					if(
 							(
+								/* #define debug(format, ...) fprintf (stderr, format, __VA_ARGS__) */
 								bIsVarArg &&
 								(paramLen == 3) &&
 								(strncmp(paramBegin[i],"...",3) == 0)
 							) || (
+								/* #define debug(MSG) fputs(MSG, stderr) */
 								(!bIsVarArg) &&
 								(paramLen == tokenLen) &&
 								(strncmp(paramBegin[i],tokenBegin,paramLen) == 0)
+							) || (
+								/* GNU cpp extension:
+								 * #define debug(format, args...) fprintf (stderr, format, args)
+								 */
+								(!bIsVarArg) &&
+								(paramLen == tokenLen + 3) &&
+								(strncmp(paramBegin[i] + tokenLen, "...", 3) == 0) &&
+								(strncmp(paramBegin[i], tokenBegin, tokenLen) == 0) &&
+								/* Let's have a side effect */
+								(bIsVarArg = true)
 							)
 						)
 					{
@@ -2594,3 +2925,17 @@ extern parserDefinition* CPreProParser (void)
 	def->useCork = CORK_QUEUE | CORK_SYMTAB;
 	return def;
 }
+
+#ifdef DEBUG
+extern void cppDebugPutc (const int level, const int c)
+{
+	if (debug (level)  &&  c != EOF)
+	{
+		     if (c == STRING_SYMBOL)  printf ("\"string\"");
+		else if (c == CHAR_SYMBOL)    printf ("'c'");
+		else                          putchar (c);
+
+		fflush (stdout);
+	}
+}
+#endif
