@@ -19,7 +19,7 @@
 
 #include "parse.h"
 #include "vstring.h"
-#include "../cpreprocessor.h"
+#include "../x-cpreprocessor.h"
 #include "debug.h"
 #include "keyword.h"
 #include "read.h"
@@ -429,6 +429,20 @@ void cxxParserSetEndLineForTagInCorkQueue(int iCorkQueueIndex,unsigned long lEnd
 {
 	CXX_DEBUG_ASSERT(iCorkQueueIndex > CORK_NIL,"The cork queue index is not valid");
 
+	tagEntryInfo *pTag = getEntryInCorkQueue (iCorkQueueIndex);
+	if (pTag)
+	{
+		/* When processing the input data in unget-buffer,
+		 * lineNumber > endLine can happen. */
+		if (pTag->lineNumber > lEndLine) {
+			CXX_DEBUG_PRINT ("Reset endline for \"%s\" starting from %ld (%ld => 0)",
+							 pTag->name,
+							 pTag->lineNumber,
+							 lEndLine);
+			lEndLine = 0;
+
+		}
+	}
 	setTagEndLineToCorkEntry (iCorkQueueIndex, lEndLine);
 }
 
@@ -438,7 +452,7 @@ void cxxParserSetEndLineForTagInCorkQueue(int iCorkQueueIndex,unsigned long lEnd
 //
 void cxxParserMarkEndLineForTagInCorkQueue(int iCorkQueueIndex)
 {
-	cxxParserSetEndLineForTagInCorkQueue(iCorkQueueIndex,getInputLineNumber());
+	cxxParserSetEndLineForTagInCorkQueue(iCorkQueueIndex, cppGetInputLineNumber());
 }
 
 
@@ -494,8 +508,8 @@ static bool cxxParserParseEnumStructClassOrUnionFullDeclarationTrailer(
 			szTypeName
 		);
 
-	MIOPos oFilePosition = getInputFilePosition();
-	int iFileLine = getInputLineNumber();
+	MIOPos oFilePosition = cppGetInputFilePosition();
+	int iFileLine = cppGetInputLineNumber();
 	int eMaybeTokenTypeOpeningBracket = (g_cxx.bConfirmedCPPLanguage
 										 ? 0
 										 : CXXTokenTypeOpeningBracket);
@@ -1999,6 +2013,8 @@ static rescanReason cxxParserMain(const unsigned int passCount)
 
 	g_cxx.iNestingLevels = 0;
 
+	g_cxx.bInCXX11Attribute = false;
+
 	bool bRet = cxxParserParseBlock(false);
 
 	cppTerminate ();
