@@ -14,7 +14,7 @@
 #include "cxx_parser_internal.h"
 
 #include "entry.h"
-#include "../cpreprocessor.h"
+#include "../x-cpreprocessor.h"
 #include "routines.h"
 #include "trashbox.h"
 #include "xtag.h"
@@ -36,16 +36,55 @@ CXX_COMMON_MACRO_ROLES(CUDA);
 	}
 
 CXX_COMMON_HEADER_ROLES(C);
-CXX_COMMON_HEADER_ROLES(CXX);
+static roleDefinition CXXHeaderRoles [] = {
+		RoleTemplateSystem,
+		RoleTemplateLocal,
+		{ true, "imported", "imported with \"imported ...\"",
+		  .version = 2 },
+		{ true, "exported", "exported with \"exported imported ...\"",
+		  .version = 2 },
+};
 CXX_COMMON_HEADER_ROLES(CUDA);
 
+/* Currently V parser wants these items. */
+#define RoleTemplateForeignDecl(V) { true, "foreigndecl", "declared in foreign languages", .version = V }
+#define RoleTemplateForeignCall(V) { true, "foreigncall", "called in foreign languages", .version = V }
 
-#define CXX_COMMON_KINDS(_langPrefix, _szMemberDescription, _syncWith)	\
+#define CXX_COMMON_FUNCTION_ROLES(__langPrefix) \
+	static roleDefinition __langPrefix##FunctionRoles [] = { \
+		RoleTemplateForeignDecl(1),							 \
+		RoleTemplateForeignCall(2),							 \
+	}
+
+CXX_COMMON_FUNCTION_ROLES(C);
+
+#define CXX_COMMON_STRUCT_ROLES(__langPrefix) \
+	static roleDefinition __langPrefix##StructRoles [] = { \
+		RoleTemplateForeignDecl(1), \
+	}
+
+CXX_COMMON_STRUCT_ROLES(C);
+
+static roleDefinition CXXModuleRoles [] = {
+
+	{ true, "partOwner", "used for specifying a partition",
+	  .version = 2 },
+	{ true, "imported", "imported with \"imported ...\"",
+	  .version = 2 },
+};
+
+static roleDefinition CXXPartitionRoles [] = {
+	{ true, "imported", "imported with \"imported ...\"",
+	  .version = 2 },
+};
+
+#define CXX_COMMON_KINDS(_langPrefix, _szMemberDescription, _syncWith, FUNC_ROLES, STRUCT_ROLES) \
 	{ true,  'd', "macro",      "macro definitions", \
 			.referenceOnly = false, ATTACH_ROLES(_langPrefix##MacroRoles), .syncWith = _syncWith \
 	}, \
 	{ true,  'e', "enumerator", "enumerators (values inside an enumeration)", .syncWith = _syncWith }, \
-	{ true,  'f', "function",   "function definitions", .syncWith = _syncWith },		\
+	{ true,  'f', "function",   "function definitions", \
+			.referenceOnly = false, FUNC_ROLES, .syncWith = _syncWith }, \
 	{ true,  'g', "enum",       "enumeration names", .syncWith = _syncWith },		\
 	{ true, 'h', "header",     "included header files", \
 			.referenceOnly = true,  ATTACH_ROLES(_langPrefix##HeaderRoles), .syncWith = _syncWith \
@@ -53,7 +92,8 @@ CXX_COMMON_HEADER_ROLES(CUDA);
 	{ false, 'l', "local",      "local variables", .syncWith = _syncWith },   \
 	{ true,  'm', "member",     _szMemberDescription, .syncWith = _syncWith },	\
 	{ false, 'p', "prototype",  "function prototypes", .syncWith = _syncWith },		\
-	{ true,  's', "struct",     "structure names", .syncWith = _syncWith },		\
+	{ true,  's', "struct",     "structure names", \
+			.referenceOnly = false, STRUCT_ROLES, .syncWith = _syncWith }, \
 	{ true,  't', "typedef",    "typedefs", .syncWith = _syncWith },			\
 	{ true,  'u', "union",      "union names", .syncWith = _syncWith },			\
 	{ true,  'v', "variable",   "variable definitions", .syncWith = _syncWith },		\
@@ -66,21 +106,27 @@ static kindDefinition g_aCXXCKinds [] = {
 	/* All other than LANG_AUTO are ignored.
 	   LANG_IGNORE is specified as a just placeholder for the macro,
 	   and is not needed. */
-	CXX_COMMON_KINDS(C,"struct, and union members", LANG_IGNORE)
+	CXX_COMMON_KINDS(C,"struct, and union members", LANG_IGNORE, ATTACH_ROLES(CFunctionRoles), ATTACH_ROLES(CStructRoles))
 };
 
 static kindDefinition g_aCXXCPPKinds [] = {
-	CXX_COMMON_KINDS(CXX,"class, struct, and union members", LANG_AUTO),
+	CXX_COMMON_KINDS(CXX,"class, struct, and union members", LANG_AUTO, .nRoles = 0, .nRoles = 0),
 	{ true,  'c', "class",      "classes" },
 	{ true,  'n', "namespace",  "namespaces" },
 	{ false, 'A', "alias",      "namespace aliases" },
 	{ false, 'N', "name",       "names imported via using scope::symbol" },
 	{ false, 'U', "using",      "using namespace statements" },
 	{ false, 'Z', "tparam",     "template parameters" },
+	{ true,  'M', "module",     "modules",
+			.referenceOnly = false, ATTACH_ROLES(CXXModuleRoles),
+			.version = 2 },
+	{ true,  'P', "partition",  "partitions",
+			.referenceOnly = false, ATTACH_ROLES(CXXPartitionRoles),
+			.version = 2 },
 };
 
 static kindDefinition g_aCXXCUDAKinds [] = {
-	CXX_COMMON_KINDS(CUDA,"struct, and union members", LANG_IGNORE)
+	CXX_COMMON_KINDS(CUDA,"struct, and union members", LANG_IGNORE, .nRoles = 0, .nRoles = 0)
 };
 
 static const char * g_aCXXAccessStrings [] = {
@@ -93,12 +139,22 @@ static const char * g_aCXXAccessStrings [] = {
 #define CXX_COMMON_FIELDS \
 	{ \
 		.name = "properties", \
-		.description = "properties (static, inline, mutable,...)", \
+		.description = "properties (static, inline, mutable, export,...)", \
 		.enabled = false \
 	}, { \
 		.name = "macrodef", \
 		.description = "macro definition", \
 			.enabled = false \
+	}, { \
+		.name = "section", \
+		.description = "the place where the object is placed", \
+		.enabled = false, \
+		.version = 1 \
+	}, { \
+		.name = "alias", \
+		.description = "the name of the alias target specified in __attribute__((alias(...)))", \
+		.enabled = false, \
+		.version = 1 \
 	}
 
 static fieldDefinition g_aCXXCFields [] = {
@@ -299,7 +355,7 @@ static short cxxTagLookBackLastNth(langType iLangType, int iScopeIndex, unsigned
 	return NO_NTH_FIELD;
 }
 
-tagEntryInfo * cxxTagBegin(unsigned int uKind,CXXToken * pToken)
+tagEntryInfo * cxxRefTagBegin(unsigned int uKind, int iRole, CXXToken * pToken)
 {
 	kindDefinition * pKindDefinitions = g_cxx.pKindDefinitions;
 
@@ -309,14 +365,14 @@ tagEntryInfo * cxxTagBegin(unsigned int uKind,CXXToken * pToken)
 		return NULL;
 	}
 
-	initTagEntry(
+	initRefTagEntry(
 			&g_oCXXTag,
 			vStringValue(pToken->pszWord),
-			uKind
+			uKind,
+			iRole
 		);
 
-	g_oCXXTag.lineNumber = pToken->iLineNumber;
-	g_oCXXTag.filePosition = pToken->oFilePosition;
+	updateTagLine (&g_oCXXTag, pToken->iLineNumber, pToken->oFilePosition);
 	g_oCXXTag.isFileScope = false;
 
 	if(!cxxScopeIsGlobal())
@@ -343,6 +399,11 @@ tagEntryInfo * cxxTagBegin(unsigned int uKind,CXXToken * pToken)
 	return &g_oCXXTag;
 }
 
+tagEntryInfo * cxxTagBegin(unsigned int uKind,CXXToken * pToken)
+{
+	return cxxRefTagBegin(uKind, ROLE_DEFINITION_INDEX, pToken);
+}
+
 vString * cxxTagSetProperties(unsigned int uProperties)
 {
 	if(uProperties == 0)
@@ -408,6 +469,8 @@ vString * cxxTagSetProperties(unsigned int uProperties)
 		ADD_PROPERTY("constinit");
 	if (uProperties & CXXTagPropertyThreadLocal)
 		ADD_PROPERTY("thread_local");
+	if (uProperties & CXXTagPropertyExport)
+		ADD_PROPERTY("export");
 
 	cxxTagSetField(CXXTagFieldProperties,vStringValue(pszProperties),false);
 
@@ -748,7 +811,7 @@ int cxxTagCommit(int *piCorkQueueIndexFQ)
 		x = vStringNewInit(g_oCXXTag.extensionFields.scopeName);
 	}
 
-	vStringCatS(x,"::");
+	vStringCatS(x, (eScopeType == CXXScopeTypeModule)? ":": "::");
 	vStringCatS(x,g_oCXXTag.name);
 
 	g_oCXXTag.name = vStringValue(x);
