@@ -24,13 +24,14 @@
 #include "parse.h"
 #include "read.h"
 #include "routines.h"
+#include "selectors.h"
 #include "vstring.h"
 #include "xtag.h"
 
 /*
 *   MACROS
 */
-#define isident(c)              (isalnum(c) || (c) == '_')
+#define isident(c)              (isalnum(c) || (c) == '_' || (c) == '$')
 #define isBlank(c)              (bool) (c == ' ' || c == '\t')
 #define isType(token,t)         (bool) ((token)->type == (t))
 #define isKeyword(token,k)      (bool) ((token)->keyword == (k))
@@ -377,6 +378,7 @@ static xtagDefinition FortranXtagTable [] = {
 		.enabled = false,
 		.name    = "linkName",
 		.description = "Linking name used in foreign languages",
+		.version = 1
 	},
 };
 
@@ -601,7 +603,7 @@ static void makeFortranLinkNameTag(tagEntryInfo *e)
 
 	tagEntryInfo ln_e = *e;
 	ln_e.name = vStringValue (ln);
-	markTagExtraBit (&ln_e, X_LINK_NAME);
+	markTagExtraBit (&ln_e, FortranXtagTable[X_LINK_NAME].xtype);
 	makeTagEntry (&ln_e);
 	vStringDelete (ln);
 }
@@ -622,8 +624,7 @@ static void makeFortranTag (tokenInfo *const token, tagType tag)
 		if (token->anonymous)
 			markTagExtraBit (&e, XTAG_ANONYMOUS);
 
-		e.lineNumber	= token->lineNumber;
-		e.filePosition	= token->filePosition;
+		updateTagLine (&e, token->lineNumber, token->filePosition);
 		e.isFileScope	= isFileScope (token->tag);
 		if (e.isFileScope)
 			markTagExtraBit (&e, XTAG_FILE_SCOPE);
@@ -2771,6 +2772,9 @@ extern parserDefinition* FortranParser (void)
 #endif
 		NULL
 	};
+	static selectLanguage selectors[] = { selectFortranOrForthByForthMarker,
+					      NULL };
+
 	parserDefinition* def = parserNew ("Fortran");
 	def->kindTable      = FortranKinds;
 	def->kindCount  = ARRAY_SIZE (FortranKinds);
@@ -2784,6 +2788,7 @@ extern parserDefinition* FortranParser (void)
 
 	def->versionCurrent = 1;
 	def->versionAge = 1;
+	def->selectLanguage = selectors;
 
 	return def;
 }
