@@ -9,6 +9,10 @@
 *
 *   This module contains code for generating tags for Windows PowerShell scripts
 *   (https://en.wikipedia.org/wiki/PowerShell).
+*
+*   References:
+*   - https://learn.microsoft.com/en-us/powershell/scripting/developer/windows-powershell-reference
+*   - https://learn.microsoft.com/en-us/powershell/scripting/lang-spec/chapter-01
 */
 
 /*
@@ -42,6 +46,7 @@ typedef enum {
 	K_CLASS,
 	K_FILTER,
 	K_ENUM,
+	K_ENUMLABEL,
 	COUNT_KIND
 } powerShellKind;
 
@@ -51,6 +56,8 @@ static kindDefinition PowerShellKinds[COUNT_KIND] = {
 	{ true, 'c', "class",		"classes" },
 	{ true, 'i', "filter",		"filter" },
 	{ true, 'g', "enum",		"enum names" },
+	{ true, 'e', "enumlabel",	"enum labels",
+	  .version = 1 },
 };
 
 
@@ -100,6 +107,7 @@ typedef struct {
 	unsigned long	lineNumber;
 	MIOPos			filePosition;
 	int				parentKind; /* KIND_GHOST_INDEX if none */
+	bool            herestring;	/* Meaningful only when type is TOKEN_STRING. */
 } tokenInfo;
 
 
@@ -124,8 +132,7 @@ static void initPowerShellEntry (tagEntryInfo *const e, const tokenInfo *const t
 {
 	initTagEntry (e, vStringValue (token->string), kind);
 
-	e->lineNumber	= token->lineNumber;
-	e->filePosition	= token->filePosition;
+	updateTagLine (e, token->lineNumber, token->filePosition);
 
 	if (access != NULL)
 		e->extensionFields.access = access;
@@ -257,6 +264,43 @@ static void parseString (vString *const string, const int delimiter)
 	}
 }
 
+static bool parseHereString (vString *const string, const int delimiter)
+{
+	enum phs_state {
+		PHS_INIT,
+		PHS_NEWLINE0,
+		PHS_QUOTE,
+		PHS_AT,
+		PHS_NEWLINE1,
+		PHS_END = PHS_NEWLINE1
+	} s = PHS_INIT;
+
+	while (s != PHS_END)
+	{
+		int c = getcFromInputFile ();
+
+		if (c == EOF)
+			return false;
+
+		vStringPut (string, c);
+
+		if (s == PHS_INIT && c == '\n')
+			s = PHS_NEWLINE0;
+		else if (s == PHS_NEWLINE0 && c == delimiter)
+			s = PHS_QUOTE;
+		else if (s == PHS_QUOTE && c == '@')
+			s = PHS_AT;
+		else if (s == PHS_AT && c == '\n')
+			s = PHS_NEWLINE1;
+		else
+			s = PHS_INIT;
+	}
+
+	Assert (vStringLength (string) >= 4);
+	vStringTruncate (string, vStringLength (string) - 4);
+	return true;
+}
+
 /* parse a identifier that may contain scopes, such as function names and
  * variable names.
  *
@@ -355,8 +399,33 @@ getNextChar:
 			parseString (token->string, c);
 			token->lineNumber = getInputLineNumber ();
 			token->filePosition = getInputFilePosition ();
+			token->herestring = false;
 			break;
 
+		case '@':
+		{
+			int d = getcFromInputFile ();
+			if (d == '\'' || d == '"')
+			{
+				bool is_herestr;
+				token->type = TOKEN_STRING;
+				is_herestr = parseHereString (token->string, d);
+				token->lineNumber = getInputLineNumber ();
+				token->filePosition = getInputFilePosition ();
+				if (is_herestr)
+					token->herestring = true;
+				else
+					token->type = TOKEN_UNDEFINED;
+			}
+			else if (d == EOF)
+				token->type = TOKEN_UNDEFINED;
+			else
+			{
+				ungetcToInputFile (d);
+				token->type = TOKEN_UNDEFINED;
+			}
+			break;
+		}
 		case '<':
 		{
 			int d = getcFromInputFile ();
@@ -649,6 +718,30 @@ static bool parseEnum (tokenInfo *const token)
 	return readNext;
 }
 
+/* parses declarations of the form
+ * 	<label> [= <int-value>]
+ * that is, contents of an enum
+ */
+static bool parseEnumLabel (tokenInfo *const token)
+{
+	bool readNext = true;
+
+	if (token->parentKind != K_ENUM)
+		return false;
+
+	if (token->type != TOKEN_IDENTIFIER)
+		return false;
+
+	makeSimplePowerShellTag (token, K_ENUMLABEL, ACCESS_UNDEFINED);
+	readToken (token);
+	if (token->type != TOKEN_EQUAL_SIGN)
+		readNext = false;
+	else /* skip int-value */
+		readToken (token);
+
+	return readNext;
+}
+
 /* parses declarations of the form
  * 	$var = VALUE
  */
@@ -736,6 +829,11 @@ static void enterScope (tokenInfo *const parentToken,
 				readNext = parseVariable (token);
 				break;
 
+			case TOKEN_IDENTIFIER:
+				if (parentKind == K_ENUM)
+					readNext = parseEnumLabel (token);
+				break;
+
 			default: break;
 		}
 
@@ -771,5 +869,9 @@ extern parserDefinition* PowerShellParser (void)
 	def->parser     = findPowerShellTags;
 	def->keywordTable = PowerShellKeywordTable;
 	def->keywordCount = ARRAY_SIZE (PowerShellKeywordTable);
+
+	def->versionCurrent = 1;
+	def->versionAge = 1;
+
 	return def;
 }
