@@ -24,6 +24,9 @@
 #include "xtag.h"
 #include "objpool.h"
 #include "ptrarray.h"
+#include "trace.h"
+
+#include "x-python.h"
 
 #define isIdentifierChar(c) \
 	(isalnum (c) || (c) == '_' || (c) >= 0x80)
@@ -76,25 +79,6 @@ static fieldDefinition PythonFields[COUNT_FIELD] = {
 	  .enabled = true },
 };
 
-typedef enum {
-	K_CLASS,
-	K_FUNCTION,
-	K_METHOD,
-	K_VARIABLE,
-	K_NAMESPACE,
-	K_MODULE,
-	K_UNKNOWN,
-	K_PARAMETER,
-	K_LOCAL_VARIABLE,
-	COUNT_KIND
-} pythonKind;
-
-typedef enum {
-	PYTHON_MODULE_IMPORTED,
-	PYTHON_MODULE_NAMESPACE,
-	PYTHON_MODULE_INDIRECTLY_IMPORTED,
-} pythonModuleRole;
-
 typedef enum {
 	PYTHON_UNKNOWN_IMPORTED,
 	PYTHON_UNKNOWN_INDIRECTLY_IMPORTED,
@@ -125,6 +109,9 @@ static roleDefinition PythonModuleRoles [] = {
 	  "namespace from where classes/variables/functions are imported" },
 	{ true, "indirectlyImported",
 	  "module imported in alternative name" },
+	{ true, "entryPoint",
+	  "specified as a module of an entry point",
+	  .version = 1 },
 };
 
 static roleDefinition PythonUnknownRoles [] = {
@@ -133,9 +120,16 @@ static roleDefinition PythonUnknownRoles [] = {
 	  "classes/variables/functions/modules imported in alternative name" },
 };
 
-static kindDefinition PythonKinds[COUNT_KIND] = {
+static roleDefinition PythonFunctionRoles [] = {
+	{ true, "entryPoint",
+	  "specified as an entry point",
+	  .version = 1 },
+};
+
+static kindDefinition PythonKinds[PYTHON_COUNT_KIND] = {
 	{true, 'c', "class",    "classes"},
-	{true, 'f', "function", "functions"},
+	{true, 'f', "function", "functions",
+	 .referenceOnly = false, ATTACH_ROLES(PythonFunctionRoles) },
 	{true, 'm', "member",   "class members"},
 	{true, 'v', "variable", "variables"},
 	{true, 'I', "namespace", "name referring a module defined in other file"},
@@ -166,7 +160,7 @@ static const keywordTable PythonKeywordTable[] = {
 };
 
 /* Taken from https://docs.python.org/3/reference/lexical_analysis.html#keywords */
-const static struct keywordGroup PythonRestKeywords = {
+static const struct keywordGroup PythonRestKeywords = {
 	.value = KEYWORD_REST,
 	.addingUnlessExisting = true,
 	.keywords = {
@@ -227,17 +221,17 @@ static accessType accessFromIdentifier (const vString *const ident,
 	const size_t len = vStringLength (ident);
 
 	/* inside a function/method, private */
-	if (parentKind != -1 && parentKind != K_CLASS)
+	if (parentKind != -1 && parentKind != PYTHON_CLASS_KIND)
 		return ACCESS_PRIVATE;
 	/* not starting with "_", public */
 	else if (len < 1 || p[0] != '_')
 		return ACCESS_PUBLIC;
 	/* "__...__": magic methods */
-	else if (kind == K_FUNCTION && parentKind == K_CLASS &&
+	else if (kind == PYTHON_FUNCTION_KIND && parentKind == PYTHON_CLASS_KIND &&
 	         len > 3 && p[1] == '_' && p[len - 2] == '_' && p[len - 1] == '_')
 		return ACCESS_PUBLIC;
 	/* "__...": name mangling */
-	else if (parentKind == K_CLASS && len > 1 && p[1] == '_')
+	else if (parentKind == PYTHON_CLASS_KIND && len > 1 && p[1] == '_')
 		return ACCESS_PRIVATE;
 	/* "_...": suggested as non-public, but easily accessible */
 	else
@@ -253,8 +247,7 @@ static void initPythonEntry (tagEntryInfo *const e, const tokenInfo *const token
 
 	initTagEntry (e, vStringValue (token->string), kind);
 
-	e->lineNumber	= token->lineNumber;
-	e->filePosition	= token->filePosition;
+	updateTagLine (e, token->lineNumber, token->filePosition);
 
 	nl = nestingLevelsGetCurrent (PythonNestingLevels);
 	if (nl)
@@ -272,8 +265,8 @@ static void initPythonEntry (tagEntryInfo *const e, const tokenInfo *const token
 			parentKind = nlEntry->kindIndex;
 
 			/* functions directly inside classes are methods, fix it up */
-			if (kind == K_FUNCTION && parentKind == K_CLASS)
-				e->kindIndex = K_METHOD;
+			if (kind == PYTHON_FUNCTION_KIND && parentKind == PYTHON_CLASS_KIND)
+				e->kindIndex = PYTHON_METHOD_KIND;
 		}
 	}
 
@@ -288,11 +281,11 @@ static int makeClassTag (const tokenInfo *const token,
                          const vString *const inheritance,
                          const vString *const decorators)
 {
-	if (PythonKinds[K_CLASS].enabled)
+	if (PythonKinds[PYTHON_CLASS_KIND].enabled)
 	{
 		tagEntryInfo e;
 
-		initPythonEntry (&e, token, K_CLASS);
+		initPythonEntry (&e, token, PYTHON_CLASS_KIND);
 
 		e.extensionFields.inheritance = inheritance ? vStringValue (inheritance) : "";
 		if (decorators && vStringLength (decorators) > 0)
@@ -311,11 +304,11 @@ static int makeFunctionTag (const tokenInfo *const token,
                             const vString *const arglist,
                             const vString *const decorators)
 {
-	if (PythonKinds[K_FUNCTION].enabled)
+	if (PythonKinds[PYTHON_FUNCTION_KIND].enabled)
 	{
 		tagEntryInfo e;
 
-		initPythonEntry (&e, token, K_FUNCTION);
+		initPythonEntry (&e, token, PYTHON_FUNCTION_KIND);
 
 		if (arglist)
 			e.extensionFields.signature = vStringValue (arglist);
@@ -357,8 +350,7 @@ static int makeSimplePythonRefTag (const tokenInfo *const token,
 		initRefTagEntry (&e, vStringValue (altName ? altName : token->string),
 		                 kind, roleIndex);
 
-		e.lineNumber	= token->lineNumber;
-		e.filePosition	= token->filePosition;
+		updateTagLine (&e, token->lineNumber, token->filePosition);
 
 		if (xtag != XTAG_UNKNOWN)
 			markTagExtraBit (&e, xtag);
@@ -465,7 +457,7 @@ static void readIdentifier (vString *const string, const int firstChar)
 	int c = firstChar;
 	do
 	{
-		vStringPut (string, (char) c);
+		vStringPut (string, c);
 		c = getcFromInputFile ();
 	}
 	while (isIdentifierChar (c));
@@ -822,7 +814,7 @@ static bool readCDefName (tokenInfo *const token, pythonKind *kind)
 
 	if (token->keyword == KEYWORD_class)
 	{
-		*kind = K_CLASS;
+		*kind = PYTHON_CLASS_KIND;
 		readToken (token);
 	}
 	else
@@ -853,7 +845,7 @@ static bool readCDefName (tokenInfo *const token, pythonKind *kind)
 				readToken (token);
 				if (token->type == '(')
 				{ /* okay, we really found a function, use this */
-					*kind = K_FUNCTION;
+					*kind = PYTHON_FUNCTION_KIND;
 					ungetToken (token);
 					copyToken (token, candidate);
 					break;
@@ -872,10 +864,12 @@ static bool readCDefName (tokenInfo *const token, pythonKind *kind)
 static vString *parseParamTypeAnnotation (tokenInfo *const token,
 										  vString *arglist)
 {
+	TRACE_ENTER();
 	readToken (token);
 	if (token->type != ':')
 	{
 		ungetToken (token);
+		TRACE_LEAVE_TEXT("type != :");
 		return NULL;
 	}
 
@@ -907,12 +901,14 @@ static vString *parseParamTypeAnnotation (tokenInfo *const token,
 							   || token->type == ',')))
 		{
 			ungetToken (token);
+			TRACE_LEAVE_TEXT("= or ,");
 			return t;
 		}
 		reprCat (arglist, token);
 		reprCat (t, token);
 	}
 	vStringDelete (t);
+	TRACE_LEAVE_TEXT("return NULL");
 	return NULL;
 }
 
@@ -986,10 +982,11 @@ static void deleteTypedParam (struct typedParam *p)
 static void parseArglist (tokenInfo *const token, const int kind,
 						  vString *const arglist, ptrArray *const parameters)
 {
+	TRACE_ENTER();
 	int prevTokenType = token->type;
 	int depth = 1;
 
-	if (kind != K_CLASS)
+	if (kind != PYTHON_CLASS_KIND)
 		reprCat (arglist, token);
 
 	do
@@ -1003,7 +1000,7 @@ static void parseArglist (tokenInfo *const token, const int kind,
 		}
 
 		readTokenFull (token, true);
-		if (kind != K_CLASS || token->type != ')' || depth > 1)
+		if (kind != PYTHON_CLASS_KIND || token->type != ')' || depth > 1)
 			reprCat (arglist, token);
 
 		if (token->type == '(' ||
@@ -1014,10 +1011,10 @@ static void parseArglist (tokenInfo *const token, const int kind,
 				 token->type == ']' ||
 				 token->type == '}')
 			depth --;
-		else if (kind != K_CLASS && depth == 1 &&
+		else if (kind != PYTHON_CLASS_KIND && depth == 1 &&
 				 token->type == TOKEN_IDENTIFIER &&
 				 (prevTokenType == '(' || prevTokenType == ',') &&
-				 PythonKinds[K_PARAMETER].enabled)
+				 PythonKinds[PYTHON_PARAMETER_KIND].enabled)
 		{
 			tokenInfo *parameterName;
 			vString *parameterType;
@@ -1032,11 +1029,13 @@ static void parseArglist (tokenInfo *const token, const int kind,
 		}
 	}
 	while (token->type != TOKEN_EOF && depth > 0);
+	TRACE_LEAVE();
 }
 
 static void parseCArglist (tokenInfo *const token, const int kind,
 						  vString *const arglist, ptrArray *const parameters)
 {
+	TRACE_ENTER();
 	int depth = 1;
 	tokenInfo *pname = newToken ();
 	vString *ptype = vStringNew ();
@@ -1142,12 +1141,14 @@ static void parseCArglist (tokenInfo *const token, const int kind,
 
 	vStringDelete (ptype);
 	deleteToken (pname);
+	TRACE_LEAVE();
 }
 
 static bool parseClassOrDef (tokenInfo *const token,
                                 const vString *const decorators,
                                 pythonKind kind, bool isCDef)
 {
+	TRACE_ENTER();
 	vString *arglist = NULL;
 	tokenInfo *name = NULL;
 	ptrArray *parameters = NULL;
@@ -1157,13 +1158,19 @@ static bool parseClassOrDef (tokenInfo *const token,
 	if (isCDef)
 	{
 		if (! readCDefName (token, &kind))
+		{
+			TRACE_LEAVE_TEXT("!readCDefName");
 			return false;
+		}
 	}
 	else
 	{
 		readToken (token);
 		if (token->type != TOKEN_IDENTIFIER)
+		{
+			TRACE_LEAVE_TEXT("token->type != TOKEN_IDENTIFIER");
 			return false;
+		}
 	}
 
 	name = newToken ();
@@ -1176,13 +1183,13 @@ static bool parseClassOrDef (tokenInfo *const token,
 		arglist = vStringNew ();
 		parameters = ptrArrayNew ((ptrArrayDeleteFunc)deleteTypedParam);
 
-		if (isCDef && kind != K_CLASS)
+		if (isCDef && kind != PYTHON_CLASS_KIND)
 			parseCArglist (token, kind, arglist, parameters);
 		else
 			parseArglist (token, kind, arglist, parameters);
 	}
 
-	if (kind == K_CLASS)
+	if (kind == PYTHON_CLASS_KIND)
 		corkIndex = makeClassTag (name, arglist, decorators);
 	else
 		corkIndex = makeFunctionTag (name, arglist, decorators);
@@ -1200,7 +1207,7 @@ static bool parseClassOrDef (tokenInfo *const token,
 		for (i = 0; i < ptrArrayCount (parameters); i++)
 		{
 			struct typedParam *parameter = ptrArrayItem (parameters, i);
-			int paramCorkIndex = makeSimplePythonTag (parameter->token, K_PARAMETER);
+			int paramCorkIndex = makeSimplePythonTag (parameter->token, PYTHON_PARAMETER_KIND);
 			tagEntryInfo *e = getEntryInCorkQueue (paramCorkIndex);
 			if (e && parameter->type)
 			{
@@ -1214,7 +1221,7 @@ static bool parseClassOrDef (tokenInfo *const token,
 
 	tagEntryInfo *e;
 	vString *t;
-	if (kind != K_CLASS
+	if (kind != PYTHON_CLASS_KIND
 		&& (e = getEntryInCorkQueue (corkIndex))
 		&& (t = parseReturnTypeAnnotation (token)))
 	{
@@ -1222,11 +1229,13 @@ static bool parseClassOrDef (tokenInfo *const token,
 		e->extensionFields.typeRef [1] = vStringDeleteUnwrap (t);
 	}
 
+	TRACE_LEAVE_TEXT("return true");
 	return true;
 }
 
 static bool parseImport (tokenInfo *const token)
 {
+	TRACE_ENTER();
 	tokenInfo *fromModule = NULL;
 
 	if (token->keyword == KEYWORD_from)
@@ -1250,7 +1259,7 @@ static bool parseImport (tokenInfo *const token)
 			/* from X import ...
 			 * --------------------
 			 * X = (kind:module, role:namespace) */
-			moduleIndex = makeSimplePythonRefTag (fromModule, NULL, K_MODULE,
+			moduleIndex = makeSimplePythonRefTag (fromModule, NULL, PYTHON_MODULE_KIND,
 												  PYTHON_MODULE_NAMESPACE,
 												  XTAG_UNKNOWN);
 		}
@@ -1288,7 +1297,7 @@ static bool parseImport (tokenInfo *const token)
 							int index;
 
 							/* Y */
-							index = makeSimplePythonRefTag (name, NULL, K_UNKNOWN,
+							index = makeSimplePythonRefTag (name, NULL, PYTHON_UNKNOWN_KIND,
 															PYTHON_UNKNOWN_INDIRECTLY_IMPORTED,
 															XTAG_UNKNOWN);
 							/* fill the scope field for Y */
@@ -1297,11 +1306,11 @@ static bool parseImport (tokenInfo *const token)
 								e->extensionFields.scopeIndex = moduleIndex;
 
 							/* Z */
-							index = makeSimplePythonTag (token, K_UNKNOWN);
+							index = makeSimplePythonTag (token, PYTHON_UNKNOWN_KIND);
 							/* fill the nameref filed for Y */
 							if (PythonFields[F_NAMEREF].enabled)
 							{
-								vString *nameref = vStringNewInit (PythonKinds [K_UNKNOWN].name);
+								vString *nameref = vStringNewInit (PythonKinds [PYTHON_UNKNOWN_KIND].name);
 								vStringPut (nameref, ':');
 								vStringCat (nameref, name->string);
 								attachParserFieldToCorkEntry (index, PythonFields[F_NAMEREF].ftype,
@@ -1316,15 +1325,15 @@ static bool parseImport (tokenInfo *const token)
 							 * x = (kind:module, role:indirectlyImported)
 							 * Y = (kind:namespace, nameref:module:x)*/
 							/* x */
-							makeSimplePythonRefTag (name, NULL, K_MODULE,
+							makeSimplePythonRefTag (name, NULL, PYTHON_MODULE_KIND,
 							                        PYTHON_MODULE_INDIRECTLY_IMPORTED,
 							                        XTAG_UNKNOWN);
 							/* Y */
-							int index = makeSimplePythonTag (token, K_NAMESPACE);
+							int index = makeSimplePythonTag (token, PYTHON_NAMESPACE_KIND);
 							/* fill the nameref filed for Y */
 							if (PythonFields[F_NAMEREF].enabled)
 							{
-								vString *nameref = vStringNewInit (PythonKinds [K_MODULE].name);
+								vString *nameref = vStringNewInit (PythonKinds [PYTHON_MODULE_KIND].name);
 								vStringPut (nameref, ':');
 								vStringCat (nameref, name->string);
 								attachParserFieldToCorkEntry (index, PythonFields[F_NAMEREF].ftype,
@@ -1346,7 +1355,7 @@ static bool parseImport (tokenInfo *const token)
 						   x = (kind:module,  role:namespace),
 						   Y = (kind:unknown, role:imported, scope:module:x) */
 						/* Y */
-						int index = makeSimplePythonRefTag (name, NULL, K_UNKNOWN,
+						int index = makeSimplePythonRefTag (name, NULL, PYTHON_UNKNOWN_KIND,
 															PYTHON_UNKNOWN_IMPORTED,
 															XTAG_UNKNOWN);
 						/* fill the scope field for Y */
@@ -1359,7 +1368,7 @@ static bool parseImport (tokenInfo *const token)
 						/* import X
 						   --------------
 						   X = (kind:module, role:imported) */
-						makeSimplePythonRefTag (name, NULL, K_MODULE,
+						makeSimplePythonRefTag (name, NULL, PYTHON_MODULE_KIND,
 						                        PYTHON_MODULE_IMPORTED,
 						                        XTAG_UNKNOWN);
 					}
@@ -1377,6 +1386,7 @@ static bool parseImport (tokenInfo *const token)
 	if (fromModule)
 		deleteToken (fromModule);
 
+	TRACE_LEAVE();
 	return false;
 }
 
@@ -1426,8 +1436,100 @@ static bool skipVariableTypeAnnotation (tokenInfo *const token, vString *const r
 	return false;
 }
 
+static bool isAtSimpleNamespace (tokenInfo *const token)
+{
+	if (token->type == TOKEN_IDENTIFIER
+		&& vStringEqC (token->string, "SimpleNamespace"))
+	{
+		tokenInfo *backup = newToken();
+
+		copyToken (backup, token);
+		readToken (token);
+		if (token->type != '(')
+		{
+			ungetToken (token);
+			copyToken (token, backup);
+			deleteToken (backup);
+			return false;
+		}
+		deleteToken (backup);
+		return true;
+	}
+
+	if (token->type == TOKEN_IDENTIFIER
+		&& vStringEqC (token->string, "types"))
+	{
+		readQualifiedName(token);
+		if (token->type == TOKEN_IDENTIFIER
+			&& vStringEqC (token->string, ".SimpleNamespace"))
+		{
+			tokenInfo *backup = newToken();
+
+			copyToken (backup, token);
+			readToken (token);
+
+			if (token->type != '(')
+			{
+				ungetToken (token);
+				copyToken (token, backup);
+				deleteToken (backup);
+				return false;
+			}
+			deleteToken (backup);
+			return true;
+		}
+	}
+	return false;
+}
+
+static void parseSimpleNamespace (tokenInfo *const token, int namespaceIndex)
+{
+	TRACE_ENTER();
+
+	nestingLevelsPush (PythonNestingLevels, namespaceIndex);
+	int prevTokenType = token->type;
+	int depth = 1;
+	int lastIndex = CORK_NIL;
+
+	do
+	{
+		if (token->type != TOKEN_WHITESPACE)
+			prevTokenType = token->type;
+
+		readTokenFull (token, true);
+
+		if (token->type == '(' ||
+			token->type == '[' ||
+			token->type == '{')
+			depth ++;
+		else if (token->type == ')' ||
+				 token->type == ']' ||
+				 token->type == '}')
+			depth --;
+		else if (depth == 1 &&
+				 token->type == TOKEN_IDENTIFIER &&
+				 (prevTokenType == '(' || prevTokenType == ','))
+			lastIndex = makeSimplePythonTag (token, PYTHON_UNKNOWN_KIND);
+		else if (depth == 1 &&
+				 token->keyword == KEYWORD_lambda &&
+				 prevTokenType == '=' &&
+				 lastIndex != CORK_NIL)
+		{
+			tagEntryInfo *e = getEntryInCorkQueue (lastIndex);
+			if (e)
+				e->kindIndex = PYTHON_FUNCTION_KIND; /* PYTHON_METHOD_KIND? */
+			lastIndex = CORK_NIL;
+		}
+	}
+	while (token->type != TOKEN_EOF && depth > 0);
+	nestingLevelsPop (PythonNestingLevels);
+
+	TRACE_LEAVE();
+}
+
 static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 {
+	TRACE_ENTER();
 	/* In order to support proper tag type for lambdas in multiple
 	 * assignations, we first collect all the names, and then try and map
 	 * an assignation to it */
@@ -1486,6 +1588,7 @@ static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 	 * we catch lambdas and alike */
 	if (token->type == '=')
 	{
+		TRACE_PRINT("operator: =");
 		unsigned int i = 0;
 
 		do
@@ -1497,19 +1600,9 @@ static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 
 			if (! nameToken)
 				/* nothing */;
-			else if (token->keyword != KEYWORD_lambda)
-			{
-				int index = makeSimplePythonTag (nameToken, kind);
-				tagEntryInfo *e = getEntryInCorkQueue (index);
-				if (e && *type)
-				{
-					e->extensionFields.typeRef [0] = eStrdup ("typename");
-					e->extensionFields.typeRef [1] = vStringDeleteUnwrap (*type);
-					*type = NULL;
-				}
-			}
-			else
+			else if (token->keyword == KEYWORD_lambda)
 			{
+				TRACE_PRINT("keyword: lambda");
 				tokenInfo *anon  = NULL;
 				vString *arglist = vStringNew ();
 				if (*type)
@@ -1564,7 +1657,7 @@ static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 					 */
 					int vindex = makeSimplePythonTag (nameToken, kind);
 					vStringClear (anon->string);
-					anonGenerate (anon->string, "anonFunc", K_FUNCTION);
+					anonGenerate (anon->string, "anonFunc", PYTHON_FUNCTION_KIND);
 					int findex = makeFunctionTag (anon, arglist, NULL);
 					tagEntryInfo *fe = getEntryInCorkQueue (findex);
 					if (fe)
@@ -1576,7 +1669,7 @@ static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 						ve->extensionFields.typeRef [0] = eStrdup ("typename");
 						ve->extensionFields.typeRef [1] = vStringDeleteUnwrap (*type);
 						*type = NULL;
-						vString *nameref = vStringNewInit (PythonKinds [K_FUNCTION].name);
+						vString *nameref = vStringNewInit (PythonKinds [PYTHON_FUNCTION_KIND].name);
 						vStringPut (nameref, ':');
 						vStringCat (nameref, anon->string);
 						attachParserField (ve, PythonFields[F_NAMEREF].ftype,
@@ -1590,6 +1683,24 @@ static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 					makeFunctionTag (nameToken, arglist, NULL);
 				vStringDelete (arglist);
 			}
+			else
+			{
+				TRACE_PRINT("name: %s", vStringValue(nameToken->string));
+				bool isAtSN = isAtSimpleNamespace(token);
+				int index = makeSimplePythonTag (nameToken,
+												 ((kind != PYTHON_LOCAL_VARIABLE_KIND) && isAtSN)
+												 ? PYTHON_NAMESPACE_KIND
+												 : kind);
+				if (isAtSN)
+					parseSimpleNamespace(token, index);
+				tagEntryInfo *e = getEntryInCorkQueue (index);
+				if (e && *type)
+				{
+					e->extensionFields.typeRef [0] = eStrdup ("typename");
+					e->extensionFields.typeRef [1] = vStringDeleteUnwrap (*type);
+					*type = NULL;
+				}
+			}
 
 			/* skip until next initializer */
 			while ((TokenContinuationDepth > 0 || token->type != ',') &&
@@ -1618,6 +1729,7 @@ static bool parseVariable (tokenInfo *const token, const pythonKind kind)
 		vStringDelete (nameTypes[nameCount]); /* NULL is acceptable. */
 	}
 
+	TRACE_LEAVE_TEXT("return false");
 	return false;
 }
 
@@ -1628,9 +1740,7 @@ static void setIndent (tokenInfo *const token)
 
 	while (lv && PY_NL (lv)->indentation >= token->indent)
 	{
-		tagEntryInfo *e = getEntryInCorkQueue (lv->corkIndex);
-		if (e)
-			e->extensionFields.endLine = token->lineNumber;
+		setTagEndLineToCorkEntry (lv->corkIndex, token->lineNumber);
 
 		nestingLevelsPop (PythonNestingLevels);
 		lv = nestingLevelsGetCurrent (PythonNestingLevels);
@@ -1639,6 +1749,8 @@ static void setIndent (tokenInfo *const token)
 
 static void findPythonTags (void)
 {
+	TRACE_ENTER();
+
 	tokenInfo *const token = newToken ();
 	vString *decorators = vStringNew ();
 	bool atStatementStart = true;
@@ -1662,14 +1774,14 @@ static void findPythonTags (void)
 		else if (token->keyword == KEYWORD_class ||
 		         token->keyword == KEYWORD_def)
 		{
-			pythonKind kind = token->keyword == KEYWORD_class ? K_CLASS : K_FUNCTION;
+			pythonKind kind = token->keyword == KEYWORD_class ? PYTHON_CLASS_KIND : PYTHON_FUNCTION_KIND;
 
 			readNext = parseClassOrDef (token, decorators, kind, false);
 		}
 		else if (token->keyword == KEYWORD_cdef ||
 		         token->keyword == KEYWORD_cpdef)
 		{
-			readNext = parseClassOrDef (token, decorators, K_FUNCTION, true);
+			readNext = parseClassOrDef (token, decorators, PYTHON_FUNCTION_KIND, true);
 		}
 		else if (token->keyword == KEYWORD_from ||
 		         token->keyword == KEYWORD_import)
@@ -1684,10 +1796,10 @@ static void findPythonTags (void)
 		{
 			NestingLevel *lv = nestingLevelsGetCurrent (PythonNestingLevels);
 			tagEntryInfo *lvEntry = getEntryOfNestingLevel (lv);
-			pythonKind kind = K_VARIABLE;
+			pythonKind kind = PYTHON_VARIABLE_KIND;
 
-			if (lvEntry && lvEntry->kindIndex != K_CLASS)
-				kind = K_LOCAL_VARIABLE;
+			if (lvEntry && lvEntry->kindIndex != PYTHON_CLASS_KIND)
+				kind = PYTHON_LOCAL_VARIABLE_KIND;
 
 			readNext = parseVariable (token, kind);
 		}
@@ -1700,9 +1812,7 @@ static void findPythonTags (void)
 				readNext = false;
 			else
 			{
-				if (vStringLength (decorators) > 0)
-					vStringPut (decorators, ',');
-				vStringCat (decorators, token->string);
+				vStringJoin(decorators, ',', token->string);
 				readToken (token);
 				readNext = skipOverPair (token, '(', ')', decorators, true);
 			}
@@ -1728,6 +1838,7 @@ static void findPythonTags (void)
 	vStringDelete (decorators);
 	deleteToken (token);
 	Assert (NextToken == NULL);
+	TRACE_LEAVE();
 }
 
 static void initialize (const langType language)
@@ -1765,5 +1876,7 @@ extern parserDefinition* PythonParser (void)
 	def->fieldCount = ARRAY_SIZE (PythonFields);
 	def->useCork = CORK_QUEUE;
 	def->requestAutomaticFQTag = true;
+	def->versionCurrent = 1;
+	def->versionAge = 1;
 	return def;
 }
