@@ -167,6 +167,10 @@ typedef enum {
 	TSTAG_ALIAS
 } tsKind;
 
+typedef enum {
+	F_PROPERTIES,
+} tsField;
+
 typedef struct sTokenInfo {
 	tokenType type;
 	keywordId keyword;
@@ -175,6 +179,7 @@ typedef struct sTokenInfo {
 	unsigned long lineNumber;
 	MIOPos filePosition;
 	keywordId accessKeyword;
+	bool isStatic;
 } tokenInfo;
 
 typedef struct sCommentState {
@@ -189,6 +194,15 @@ typedef struct sBlockState {
 	int curlyLevel;
 } blockState;
 
+static fieldDefinition TsFields[] = {
+	{
+		.name = "properties",
+		.description = "properties (static)",
+		.enabled = false,
+		.version = 1,
+	},
+};
+
 static const keywordTable TsKeywordTable [] = {
 	/* keyword		  keyword ID */
 	{ "as"	        , KEYWORD_as          },
@@ -278,8 +292,7 @@ static int emitTag(const tokenInfo *const token, const tsKind kind)
 	tagEntryInfo e;
 
 	initTagEntry (&e, name, kind);
-	e.lineNumber   = token->lineNumber;
-	e.filePosition = token->filePosition;
+	updateTagLine (&e, token->lineNumber, token->filePosition);
 	e.extensionFields.scopeIndex = token->scope;
 
 	switch (token->accessKeyword)
@@ -295,6 +308,9 @@ static int emitTag(const tokenInfo *const token, const tsKind kind)
 			break;
 	}
 
+	if (token->isStatic)
+		attachParserField (&e, TsFields[F_PROPERTIES].ftype, "static");
+
 	return makeTagEntry (&e);
 }
 
@@ -319,6 +335,7 @@ static void clearPoolToken (void *data)
 	token->scope = CORK_NIL;
 
 	token->accessKeyword = KEYWORD_NONE;
+	token->isStatic      = false;
 
 	token->string = vStringNewOrClear (token->string);
 }
@@ -1415,7 +1432,6 @@ static void parseFunctionBody (const int scope, tokenInfo *const token)
 								parseStringDQuote,
 								parseStringTemplate,
 								parseStringRegex,
-								parseTemplate,
 								parseVarKeyword,
 								parseLetKeyword,
 								parseConstKeyword,
@@ -1664,11 +1680,7 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 		if (token->type == TOKEN_KEYWORD
 			&& (token->keyword == KEYWORD_extends || token->keyword == KEYWORD_implements)
 			&& inheritance == NULL) inheritance = vStringNew ();
-		else if (inheritance && token->type == TOKEN_IDENTIFIER)
-		{
-			if (!vStringIsEmpty (inheritance)) vStringPut(inheritance, ',');
-			vStringCat(inheritance, token->string);
-		}
+		else if (inheritance && token->type == TOKEN_IDENTIFIER) vStringJoin (inheritance, ',', token->string);
 	} while (parsed && token->type != TOKEN_OPEN_CURLY);
 
 	if (! parsed)
@@ -1692,6 +1704,7 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 	bool isGenerator = false;
 	bool parsingValue = false;
 	int visibility = 0;
+	bool isStatic = false;
 
 	do
 	{
@@ -1768,6 +1781,16 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 							visibility = token->keyword;
 							parsingValue = false;
 							break;
+						case KEYWORD_static:
+							if (member)
+							{
+								emitTag (member, TSTAG_PROPERTY);
+								deleteToken (member);
+								member = NULL;
+							}
+							isStatic = true;
+							parsingValue = false;
+							break;
 						case KEYWORD_new:
 						case KEYWORD_typeof:
 							parsingValue = true;
@@ -1796,6 +1819,7 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 						member = NULL;
 						isGenerator = false;
 						visibility = 0;
+						isStatic = false;
 					}
 					parsingValue = false;
 					break;
@@ -1836,6 +1860,7 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 
 					isGenerator = false;
 					visibility = 0;
+					isStatic = false;
 					parsingValue = false;
 					break;
 				case TOKEN_IDENTIFIER:
@@ -1844,6 +1869,7 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 						member = newToken ();
 						copyToken (member, token, false);
 						member->scope = scope;
+						member->isStatic = isStatic;
 						if (visibility) member->accessKeyword = visibility;
 						else member->accessKeyword = KEYWORD_public;
 					}
@@ -1854,6 +1880,7 @@ static void parseClassBody (const int scope, tokenInfo *const token)
 				default:
 					isGenerator = false;
 					visibility = 0;
+					isStatic = false;
 					break;
 			}
 		}
@@ -2127,6 +2154,8 @@ extern parserDefinition *TypeScriptParser (void)
 	def->extensions = extensions;
 	def->kindTable  = TsKinds;
 	def->kindCount  = ARRAY_SIZE (TsKinds);
+	def->fieldTable = TsFields;
+	def->fieldCount = ARRAY_SIZE (TsFields);
 	def->parser     = findTsTags;
 	def->initialize = initialize;
 	def->finalize   = finalize;
@@ -2138,5 +2167,8 @@ extern parserDefinition *TypeScriptParser (void)
 	def->initStats = initStats;
 	def->printStats = printStats;
 
+	def->versionCurrent = 1;
+	def->versionAge = 1;
+
 	return def;
 }
